# -*- coding: utf-8 -*-
# PIL and pyfits are show-stoppers for this plugin
try:
  import PIL.Image
except:
# image.py will complain about this one too, more verbosely
  print """Python Imaging Library (PIL) not found, rendering of FITS files will not be available.
PIL can be installed from the Debian/Ubuntu package python-imaging, or can be downloaded from
http://www.pythonware.com/index.htm.
""";
  raise;

try:
  ## ugly hack to get around UGLY FSCKING ARROGNAT (misspelling fully intentional) pyfits-2.3 bug
  import Kittens.utils
  pyfits = Kittens.utils.import_pyfits();
except:
  print """PyFITS package not found, rendering of FITS files will not be available.
PyFITS can be installed from the Debian/Ubuntu package python-pyfits, or can be downloaded
from http://www.stsci.edu/resources/software_hardware/pyfits.
""";
  raise

try:
  import numpy
  import scipy.ndimage.measurements
except:
  print """numpy and/or scipy packages not found, rendering of FITS files will not be available.
Numpy and scipy can be installed from Debian/Ubuntu packages python-numpy and python-scipy, or can
be downloaded from http://numpy.scipy.org.
""";
  raise

# pychart needed for rendering histograms, but we can get on without it
from local_pychart import *
pychart = True;

#except:
#  pychart = None;
#  print """PyChart package not found, rendering of FITS histograms will not be available.
#PyChart is available from http://home.gna.org/pychart/, or as Debian package python-pychart.
#""";

import os.path
import traceback
import math
import cPickle
import time

import Purr
from Purr.Render import quote_url,dprint,dprintf
from Purr.CachingRenderer import CachingRenderer
import Kittens.utils

class FITSRenderer (CachingRenderer):
  """This class renders FITS image data products.""";
  def canRender (filename):
    """We can render it if PIL can read it.""";
    if filename.endswith(".fits") or filename.endswith(".FITS"):
      return 10;
    return False;
  canRender = staticmethod(canRender);

  # this gives a short ID for the class (used in GUIs and such)
  renderer_id = "fits";

  # this gives a documentation string. You can use rich text here
  renderer_doc = """<P>The "fits" plugin provides rendering of FITS images.""";

  # maximum thumbnail width & height
  # define renderer options
  CachingRenderer.addOption("image-thumbnail-width",512,dtype=int,doc="Maximum width of thumbnails");
  CachingRenderer.addOption("image-thumbnail-height",256,dtype=int,doc="Maximum height of thumbnails");
  CachingRenderer.addOption("hist-width",1024,dtype=int,doc="Width of histogram plot");
  CachingRenderer.addOption("hist-height",1024,dtype=int,doc="Height of histogram plot");
  CachingRenderer.addOption("hist-thumbnail-width",128,dtype=int,doc="Maximum width of thumbnails");
  CachingRenderer.addOption("hist-thumbnail-height",128,dtype=int,doc="Maximum height of thumbnails");
  CachingRenderer.addOption("fits-nimage",4,dtype=int,doc="Maximum number of planes to include, when dealing with cubes");
  CachingRenderer.addOption("fits-hist-nbin",512,dtype=int,doc="Number of bins to use when making histograms""");
  CachingRenderer.addOption("fits-hist-clip",.95,dtype=float,doc="Apply histogram clipping");

  def _renderCacheFile (self,cachetype,relpath):
    return ("-cache-%s-rel-%s.html"%(cachetype,bool(relpath))).lower();

  @staticmethod
  def compute_tics (x0,x1):
    # work out tic interval
    dx = x1-x0;
    tic = 10**math.floor(math.log10(dx));
    if dx/tic < 2:
      tic /= 5;
    elif dx/tic < 5:
      tic /= 2;
    # work out where the tics are (in terms of i*interval)
    it0 = int(math.floor(x0/tic));
    it1 = int(math.floor(x1/tic));
    if it0*tic < x0:
      it0 += 1;
    # scale back up and return
    return [ i*tic for i in range(it0,it1+1) ];

  def _make_histogram (self,path,title,x,y):
    # make histogram by doubling up each x point to make "horizontals"
    x = map(float,x);
    y = map(int,y);
    width = x[1]-x[0];
    xy = [];
    for a,b in zip(x,y):
      xy += [(a,b),(a+width,b)];
    # Make sure all spaces in path are escaped as pychart cannot handle them
    path = path.replace('\ ', ' ').replace(' ', '\ ');
    canv = canvas.init(path);
    ar = area.T(
      x_axis=axis.X(label="/20{}"+title,format="/20{}%g",tic_interval=self.compute_tics),
      y_axis=axis.Y(label=None,format="/20{}%s"),
      x_grid_style=line_style.gray70,
      y_grid_style=line_style.gray70,
      legend=None,
      size=self.hist_size
    );
    ar.add_plot(line_plot.T(
      data=xy,xcol=0,ycol=1,
      line_style=line_style.black,
      label=None,
      data_label_format=None
    ));
    ar.draw(canv);
    canv.close();

  def regenerate (self):
    Purr.progressMessage("reading %s"%self.dp.filename,sub=True);
    # init fitsfile to None, so that _read() above is forced to re-read it
    fitsfile = pyfits.open(self.dp.fullpath);
    header = fitsfile[0].header;

    dprintf(3,"beginning render of",self.dp.fullpath); t0 = time.time();
    # write out FITS header
    self.headerfile,path,uptodate = self.subproduct("-fitsheader.html");
    if not uptodate:
      title = "FITS header for %s"%self.dp.filename;
      html = """<HTML><BODY><TITLE>%s</TITLE>
      <H2>%s</H2>
      <PRE>"""%(title,title);
      for line in header.ascard:
        line = str(line).replace("<","&lt;").replace(">","&gt;");
        html += line+"\n";
      html += """
      </PRE></BODY></HTML>\n""";
      try:
        file(path,"w").write(html);
      except:
        print "Error writing file %s"%path;
        traceback.print_exc();
        self.headerfile = None;

    # figure out number of images to include
    ndim = header['NAXIS'];
    fitsshape = [ header['NAXIS%d'%i] for i in range(1,ndim+1) ];
    self.cubesize = 'x'.join(map(str,fitsshape));
    if ndim < 2:
      raise TypeError,"can't render one-dimensional FITS files""";
    elif ndim == 2:
      fitsdata_to_images = lambda fdata:[fdata];
      nplanes = 1;
    else:
      ax1 = ax2 = None;
      # find the X/Y axes, by looking at CTYPEx
      # note that the array axes are in reverse order. I.e. if X is FITS axis 1 and Y is axis 2,
      # the array will be of e.g. shape 1,1,NY,NX, while fitsshape is [NX,NY,1,1]
      for i in range(1,ndim+1):
        ctype = header['CTYPE%d'%i];
        if [ prefix for prefix in "RA","GLON","ELON","HLON","SLON" if ctype.startswith(prefix) ] \
            or ctype in ("L","X"):
          ax1 = ndim-i;
        elif [ prefix for prefix in "DEC","GLAT","ELAT","HLAT","SLAT" if ctype.startswith(prefix) ] \
            or ctype in ("M","Y"):
          ax2 = ndim-i;
      if ax1 is None or ax2 is None:
        ax1,ax2 = 1,0;
      arrshape = fitsshape[-1::-1];
      # this is how many planes we render, at most
      nplanes = max(self.getOption('fits-nimage'),1);
      slices = [];
      baseslice = [0]*ndim;
      baseslice[ax1] = baseslice[ax2] = None;
      imgshape = (arrshape[min(ax1,ax2)],arrshape[max(ax1,ax2)]);
      while len(slices) < nplanes:
        slices.append(tuple(baseslice));
        for idim in range(ndim):
          if baseslice[idim] != None:
            baseslice[idim] += 1;
            if baseslice[idim] < arrshape[idim]:
              break;
            else:
              baseslice[idim] = 0;
        else:
          break;
      nplanes = len(slices);
      # OK, slices contains how many slices to return
      def fitsdata_to_images (fdata,slices=slices,imgshape=imgshape):
        dprint(3,"fitsdata_to_images",slices,fdata.shape); t0 = time.time();
        # reshape to collapse into a 3D cube
        img = [ fdata[i].reshape(imgshape) for i in slices ];
        dprint(3,"collecting images took",time.time()-t0,"secs"); t0 = time.time();
        return img;

    # OK, now cycle over all images
    dprintf(3,"%s: rendering %d planes\n",self.dp.fullpath,nplanes); t0 = time.time();

    self.imgrec = [None]*nplanes;
    # get number of bins (0 or None means no histogram)
    nbins = self.getOption("fits-hist-nbin");
    # see if histogram clipping is enabled, set hclip to None if not
    self.hclip = hclip = self.getOption("fits-hist-clip");
    if hclip == 1 or not nbins:
      hclip = None;

    tsize_img  = self.getOption("image-thumbnail-width"),self.getOption("image-thumbnail-height");
    tsize_hist = self.getOption("hist-thumbnail-width"),self.getOption("hist-thumbnail-height");
    self.hist_size = self.getOption("hist-width"),self.getOption("hist-height");

    # filled once we read the data
    images = None;

    for num_image in range(nplanes):
      # do we have a cached status record for this image?
      recfile,recpath,uptodate = self.subproduct("-%d-stats"%num_image);
      if uptodate:
        dprintf(3,"%s(%d): stats file %s up-to-date, reading in\n",self.dp.fullpath,num_image,recfile);
        try:
          self.imgrec[num_image] = cPickle.load(file(recpath));
          continue;
        except:
          print "Error reading stats file %s, regenerating everything"%recpath;
          traceback.print_exc();
      # out of date, so we regenerate everything
      # build up record of stuff associated with this image
      rec = self.imgrec[num_image] = Kittens.utils.recdict();

      # generate paths for images
      rec.fullimage,img_path  = self.subproductPath("-%d-full.png"%num_image);
      rec.thumbnail,img_thumb = self.subproductPath("-%d-thumb.png"%num_image);
      if pychart:
        rec.histogram_full,hf_path = self.subproductPath("-%d-hist-full.png"%num_image);
        rec.histogram_zoom,hz_path = self.subproductPath("-%d-hist-zoom.png"%num_image);
        rec.histogram_full_thumb,hf_thumb = self.subproductPath("-%d-hist-full-thumb.png"%num_image);
        rec.histogram_zoom_thumb,hz_thumb = self.subproductPath("-%d-hist-zoom-thumb.png"%num_image);

      # need to read in data at last
      if not images:
        dprint(3,"reading data"); t0 = time.time();
        fitsdata = fitsfile[0].data;
        dprint(3,"reading data took",time.time()-t0,"secs"); t0 = time.time();
        fitsfile = None;
        images = fitsdata_to_images(fitsdata);
        dprint(3,"converting to images took",time.time()-t0,"secs"); t0 = time.time();
        fitsdata = None;

      data = images[num_image];

      title = self.dp.filename;
      if nplanes > 1:
        title += ", plane #%d"%num_image;
      Purr.progressMessage("rendering %s"%title,sub=True);

      # min/max data values
      dprint(3,"rendering plane",num_image); t0 = time.time();
      datamask = ~numpy.isfinite(data);
      dprint(3,"making mask took",time.time()-t0,"secs"); t0 = time.time();
      datamin,datamax = scipy.ndimage.measurements.extrema(data,datamask,False)[:2];
      dprint(3,"computing min/max took",time.time()-t0,"secs"); t0 = time.time();
      rec.datamin,rec.datamax = datamin,datamax;
      # mean and sigma
      rec.datamean = scipy.ndimage.measurements.mean(data,datamask,False);
      dprint(3,"computing mean took",time.time()-t0,"secs"); t0 = time.time();
      rec.datastd = scipy.ndimage.measurements.standard_deviation(data,datamask,False);
      dprint(3,"computing std took",time.time()-t0,"secs"); t0 = time.time();
      # thumbnail files will be "" if images are small enough to be inlined.
      # these will be None if no histogram clipping is applied
      rec.clipmin,rec.clipmax = None,None;
      dprintf(3,"%s plane %d: datamin %g, datamax %g\n",self.dp.fullpath,num_image,rec.datamin,rec.datamax);
      # compute histogram of data only if this is enabled,
      # and either pychart is available (so we can produce plots), or histogram clipping is in effect
      if datamin != datamax and nbins and (pychart or hclip):
        dprintf(3,"%s plane %d: computing histogram\n",self.dp.fullpath,num_image);
        counts = scipy.ndimage.measurements.histogram(data,datamin,datamax,nbins,labels=datamask,index=False); # needed for 1.3+ to avoid warnings
        edges = datamin + (datamax-datamin)*(numpy.arange(nbins,dtype=float)+.5)/nbins;
        dprint(3,"computing histogram took",time.time()-t0,"secs"); t0 = time.time();
        # render histogram
        if pychart:
          try:
            self._make_histogram(hf_path,"Histogram of %s"%title,edges,counts);
            dprint(3,"rendering histogram took",time.time()-t0,"secs"); t0 = time.time();
          except:
            print "Error rendering histogram %s"%hf_path;
            traceback.print_exc();
            rec.histogram_full = None;
          # if histogram was rendered, make a thumbnail
          if rec.histogram_full:
            self.makeThumb(hf_path,hf_thumb,tsize_hist);
          else:
            rec.histogram_full_thumb = None;
        # now, compute clipped data if needed
        if hclip:
          # find max point in histogram
          ic = counts.argmax();
          # compute number of points that need to be included, given the clip factor
          target_count = int(data.size*hclip);
          ih0 = ih1 = ic;
          totcount = counts[ic];
          # find how many bins to include around ic, stopping when we hit the edge
          while totcount < target_count:
            if ih0 > 0:
              ih0 -= 1;
              totcount += counts[ih0];
            if ih1 < nbins-1:
              ih1 += 1;
              totcount += counts[ih1];
            # just in case
            if ih0 <= 0 and ih1 >= nbins-1:
              break;
          # and these are the clipping limits
          datamin = float(edges[ih0])
          if ih1 >= nbins-1:
            ih1 = nbins-1;  # and datamax is already the clipping limit
          else:
            ih1 += 1;
            datamax = float(edges[ih1]);
          rec.clipmin,rec.clipmax = datamin,datamax;
          dprintf(3,"%s plane %d: clipping to %g,%g\n",self.dp.fullpath,num_image,rec.clipmin,rec.clipmax);
          # render zoomed histogram
          if pychart:
            if rec.clipmax != rec.clipmin:
              zcounts = scipy.ndimage.measurements.histogram(data,rec.clipmin,rec.clipmax,nbins,labels=datamask,index=False); # needed for 1.3+ to avoid warnings
              zedges = rec.clipmin + (rec.clipmax-rec.clipmin)*(numpy.arange(nbins,dtype=float)+.5)/nbins;
              try:
                self._make_histogram(hz_path,"Histogram zoom of %s"%title,zedges,zcounts);
                dprint(3,"rendering zoomed histogram took",time.time()-t0,"secs"); t0 = time.time();
              except:
                print "Error rendering histogram %s"%hz_path;
                traceback.print_exc();
                rec.histogram_zoom = None;
            else: # no meaningful zoomed area to render
              rec.histogram_zoom = None;
            # if histogram was rendered, make a thumbnail
            if rec.histogram_zoom:
              histogram_zoom_thumb = self.makeThumb(hz_path,hz_thumb,tsize_hist);
            else:
              rec.histogram_zoom_thumb = None;
          # clip data
          data = numpy.clip(data,datamin,datamax);
        # end of clipping
      # else no histogram for whatever reason
      else:
        rec.histogram_full = rec.histogram_zoom = rec.histogram_full_thumb = rec.histogram_zoom_thumb = None;
      # ok, data has been clipped if need be. Rescale it to 8-bit integers
      t0 = time.time();
      datarng = datamax - datamin;
      if datarng:
        data = (data - datamin)*(255/datarng)
        data = data.round().astype('uint8');
        data[datamask] = 255;
      else:
        data = numpy.zeros(data.shape,dtype='uint8');
      dprintf(3,"%s plane %d: rescaled to %d:%d in %f seconds\n",self.dp.fullpath,num_image,data.min(),data.max(),time.time()-t0); t0 = time.time();
      # generate PNG image
      img = None;
      try:
        img = PIL.Image.frombuffer('L',data.shape[-1::-1],numpy.getbuffer(data),"raw",'L',0,-1);
        dprint(3,"image frombuffer took",time.time()-t0,"secs"); t0 = time.time();
        #img = PIL.Image.new('L',data.shape);
        #dprint(3,"new image took",time.time()-t0,"secs"); t0 = time.time();
        #imgdata = data.reshape((data.size,));
        #dprint(3,"data.reshape took",time.time()-t0,"secs"); t0 = time.time();
        #img.putdata(imgdata);
        #dprint(3,"putdata took",time.time()-t0,"secs"); t0 = time.time();
        # img = img.transpose(PIL.Image.FLIP_TOP_BOTTOM);
        # dprint(3,"transpose took",time.time()-t0,"secs"); t0 = time.time();
        img.save(img_path,'PNG');
        dprint(3,"saving took",time.time()-t0,"secs"); t0 = time.time();
      except:
        print "Error rendering image %s"%path;
        traceback.print_exc();
        rec.fullimage = img = None;
      # if image was rendered, make a thumbnail
      if rec.fullimage:
        thumb = self.makeThumb(img_path,img_thumb,tsize_img,img=img);
        dprint(3,"rendering thumbnail took",time.time()-t0,"secs"); t0 = time.time();
        # None means thumbnail failed
        if thumb is None:
          rec.thumbnail = None;
        # else perhaps image is its own thumbnail
        elif thumb is img_path:
          rec.thumbnail = rec.fullimage;
      else:
        rec.thumbnail = None;
      # write stats
      try:
        cPickle.dump(rec,file(recpath,'w'));
      except:
        print "Error writing stats file  %s"%recpath;
        traceback.print_exc();

  def makeThumb (self,imagepath,thumbpath,tsize,img=None):
    """makes a thumbnail for the given image.
    imagepath refers to an image file
    img can be an open PIL.Image -- if None, then imagepath is opened
    tsize is a width,height tuple giving the max thumbnail size
    extension is the extension of the thumbnail file
    """
    try:
      # open image if needed
      if not img:
        img = PIL.Image.open(imagepath);
      # do we need a thumbnail at all, or can the image be inlined?
      width,height = img.size;
      factor = max(width/float(tsize[0]),height/float(tsize[1]));
      if factor <= 1:
        return imagepath;
      # generate the thumbnail
      img = img.resize((int(width/factor),int(height/factor)),PIL.Image.ANTIALIAS);
      img.save(thumbpath,"PNG");
      return thumbpath;
    except:
      print "Error rendering thumbnail %s"%thumbpath;
      traceback.print_exc();
      return None;

  def _renderSingleImage (self,image,thumb,relpath):
    if image is None:
      return "";
    # else no thumbnail: make "image" link
    elif thumb is None:
      fname = relpath+image;
      return """<A HREF="%s">image</A>"""%quote_url(fname);
    # else return thumbnail linking to full image
    else:
      tname = relpath+thumb;
      fname = relpath+image;
      return """<A HREF="%s"><IMG SRC="%s" ALT="%s"></A>"""%(quote_url(fname),quote_url(tname),
                                                             quote_url(os.path.basename(image)));

  def _renderImageRec (self,rec,relpath,include_size=False):
    # get HTML code for image and histograms
    html_image = self._renderSingleImage(rec.fullimage,rec.thumbnail,relpath);
    if rec.histogram_full:
      html_hist_full = self._renderSingleImage(rec.histogram_full,rec.histogram_full_thumb,relpath);
    else:
      html_hist_full = "";
    if rec.histogram_zoom:
      html_hist_zoom = self._renderSingleImage(rec.histogram_zoom,rec.histogram_zoom_thumb,relpath);
    else:
      html_hist_zoom = "";
    # arrange images side-by-side
    html_img = "<TABLE><TR><TD ROWSPAN=2>%s</TD><TD>%s</TD></TR><TR><TD>%s</TD></TR></TABLE>"% \
                      (html_image,html_hist_full,html_hist_zoom);
    # form up comments
    html_cmt = """<TABLE><TR><TD>data range:</TD><TD>%g,%g</TD></TR>"""%(rec.datamin,rec.datamax);
    if include_size:
      html_cmt += """
         <TR><TD>size:</TD><TD>%s</TD></TR>\n"""%self.cubesize;
    html_cmt += """
         <TR><TD>mean:</TD><TD>%g</TD></TR>
         <TR><TD>sigma:</TD><TD>%g</TD></TR>"""%(rec.datamean,rec.datastd);
    if rec.clipmin is not None:
      html_cmt += """
         <TR><TD>clipping:</TD><TD>%g%%</TD></TR>
         <TR><TD>clip range:</TD><TD>%g,%g</TD></TR>"""%(self.hclip*100,rec.clipmin,rec.clipmax);
    html_cmt += """\n
       </TABLE>""";
    return html_img,html_cmt;

  def renderLink (self,relpath=""):
    """renderLink() is called to render a link to the DP
    """;
    # return from cache if available
    cachekey,html = self.checkCache('Link',relpath);
    if html is not None:
      return html;
    # else regenerate
    html = CachingRenderer.renderLink(self,relpath);
    if self.headerfile is not None:
      html += """ (<A HREF="%s">header</A>)"""%quote_url(relpath+self.headerfile);
    # save to cache
    return self.writeCache(cachekey,html);


  def renderInTable (self,relpath=""):
    """renderInTable() is called to render FITS images in a table""";
    # return from cache if available
    cachekey,html = self.checkCache('InTable',relpath);
    if html is not None:
      return html;
    # else regenerate
    # single image: render as standard cells
    if len(self.imgrec) == 1:
      rec = self.imgrec[0];
      # add header
      html = "    <TR><TD COLSPAN=2>";
      html += self.renderLinkComment(relpath) or "";
      html += "</TD></TR>\n";
      html_img,comment = self._renderImageRec(rec,relpath,include_size=True);
      html += "\n".join([
          "    <TR>",
          "      <TD>%s</TD>"%html_img,
          "      <TD>%s</TD>"%comment,
          "    </TR>\n" ]);
    # multiple images: render a single header row, followed by one row per image
    else:
      # add header
      html = "    <TR><TD COLSPAN=2>";
      html += self.renderLinkComment(relpath);
      # append information on image and end the table row
      html += "\n      <DIV ALIGN=right><P>%s FITS cube, %d planes are given below.</P></DIV></TD></TR>\n"%(self.cubesize,len(self.imgrec));
      # now loop over images and generate a table row for each
      for irec,rec in enumerate(self.imgrec):
        html_img,comment = self._renderImageRec(rec,relpath);
        comment = "<P>Image plane #%d.</P>%s"%(irec,comment);
        html += "\n".join([
            "    <TR>" ,
            "      <TD>%s</TD>"%html_img,
            "      <TD>%s</TD>"%comment,
            "    </TR>\n" ]);
    return self.writeCache(cachekey,html);

  def renderThumbnail (self,relpath=""):
    """renderThumbnail() is called to render a thumbnail of the DP.
    We only render the first image (in case of multiple images)
    """;
    # return from cache if available
    cachekey,html = self.checkCache('Thumbnail',relpath);
    if html is not None:
      return html;
    # else regenerate
    rec = self.imgrec[0];
    html = self._renderSingleImage(rec.fullimage,rec.thumbnail,relpath);
    # save to cache
    return self.writeCache(cachekey,html);

# register ourselves with Purr
import Purr.Render
Purr.Render.addRenderer(FITSRenderer,__name__,__file__);


# -*- coding: utf-8 -*-
from PyQt4.Qt import *

import time
import os
import os.path
import string

from Purr import Config,pixmaps,dprint,dprintf
import Purr.LogEntry
import Purr.Render
import Kittens.widgets
import Kittens.utils

def _makeUniqueFilename (taken_names,name):
  """Helper function. Checks if name is in the set 'taken_names'.
  If so, attepts to form up an untaken name (by adding numbered suffixes).
  Adds name to taken_names.
  """
  if name in taken_names:
    # try to form up new name
    basename,ext = os.path.splitext(name);
    num = 1;
    name = "%s-%d%s"%(basename,num,ext);
    while name in taken_names:
      num += 1;
      name = "%s-%d%s"%(basename,num,ext);
  # finally, enter name into set
  taken_names.add(name);
  return name;


_sanitize_chars = "\\:*?\"<>|";
_sanitize_trans = string.maketrans(_sanitize_chars,'_'*len(_sanitize_chars));

def _sanitizeFilename (filename):
  """Sanitizes filename for use on Windows and other brain-dead systems, by replacing a number of illegal characters
  with underscores."""
  global _sanitize_trans;
  out = filename.translate(_sanitize_trans);
  # leading dot becomes "_"
  if out and out[0] == '.':
    out = out[1:];
  return out;

class DPTreeWidget (Kittens.widgets.ClickableTreeWidget):
  """This class implements a QTreeWidget for data products.
  """;
  def __init__ (self,*args):
    Kittens.widgets.ClickableTreeWidget.__init__(self,*args);
    # insert columns, and numbers for them
    self.header().setDefaultSectionSize(120);
    columns = [ "action","filename","type","rename to","render","comment" ];
    self.setHeaderLabels(columns);
    self.ColAction,self.ColFilename,self.ColType,self.ColRename,self.ColRender,self.ColComment = range(len(columns));
    self.setSortingEnabled(False);
    self.setRootIsDecorated(False);
    self.setEditTriggers(QAbstractItemView.AllEditTriggers);
    self.setDragEnabled(True);
    # sort out resizing modes
    self.header().setMovable(False);
    self.header().setResizeMode(self.ColFilename,QHeaderView.Interactive);
    self.header().setResizeMode(self.ColType,QHeaderView.ResizeToContents);
    self.header().setResizeMode(self.ColRename,QHeaderView.Interactive);
    self.header().setResizeMode(self.ColComment,QHeaderView.Stretch);
    self.header().setResizeMode(self.ColAction,QHeaderView.Fixed);
    self.header().setResizeMode(self.ColRender,QHeaderView.Fixed);
    self.header().resizeSection(self.ColAction,64);
    self.header().resizeSection(self.ColRender,64);
    self._fontmetrics = QFontMetrics(QFont());
    # setup other properties of the listview
    self.setAcceptDrops(True);
    try: self.setAllColumnsShowFocus(True);
    except AttributeError: pass; # qt 4.2+
    # self.setDefaultRenameAction(QTreeWidget.Accept);
    self.setSelectionMode(QTreeWidget.SingleSelection);
    self.header().show();
    self.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding);
    self.connect(self,SIGNAL("mouseButtonClicked"),self._itemClicked);
    self.connect(self,SIGNAL("itemActivated(QTreeWidgetItem*,int)"),self._itemActivated);
    self.connect(self,SIGNAL("itemChanged(QTreeWidgetItem*,int)"),self._itemRenamed);
    self.connect(self,SIGNAL("currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)"),self._currentItemChanged);
    self.connect(self,SIGNAL("itemcontextMenuRequested"),self._showItemContextMenu);
    self.connect(self,SIGNAL("droppedFiles"),self,SIGNAL("filesSelected"));
    # create popup menu for existing DPs
    self._archived_dp_menu = menu = QMenu();
    menu.addAction(pixmaps.editcopy.icon(),"Restore file from this entry's archived copy",self._restoreItemFromArchive);
    menu.addAction(pixmaps.editpaste.icon(),"Copy location of archived copy to clipboard",self._copyItemToClipboard);
    # currently selected item
    self._current_item = None;
    # item onto which something was dropped
    self._dropped_on = None;
    # currently edited item and column
    self._editing = None;
    # dictionary of listview items
    # key is (path,archived) tuple, value is item
    self.dpitems = {};

    self._policy_list_default         = ( ("copy",pixmaps.copy.icon()),
                                    ("move",pixmaps.move.icon()),
                                    ("ignore",pixmaps.grey_round_cross.icon()),
                                    ("banish",pixmaps.red_round_cross.icon()) );
    self._policy_list_archived         = ( ("keep",pixmaps.checkmark.icon()),
                                    ("remove",pixmaps.grey_round_cross.icon()) );

  def _checkDragDropEvent (self,ev):
    """Checks if event contains a file URL, accepts if it does, ignores if it doesn't""";
    mimedata = ev.mimeData();
    if mimedata.hasUrls():
      urls = [ str(url.toLocalFile()) for url in mimedata.urls() if url.toLocalFile() ];
    else:
      urls = [];
    # accept event if drag text is a file URL
    if urls:
      ev.acceptProposedAction();
      return urls;
    else:
      ev.ignore();
      return None;

  def dragEnterEvent (self,ev):
    """Process drag-enter event. Use function above to accept or ignore it"""
    self._checkDragDropEvent(ev);

  def dragMoveEvent (self,ev):
    """Process drag-move event. Use function above to accept or ignore it"""
    self._checkDragDropEvent(ev);

  def dropEvent (self,ev):
    """Process drop event."""
    # use function above to accept event if it contains a file URL
    files = self._checkDragDropEvent(ev);
    if files:
      pos = ev.pos();
      dropitem = self.itemAt(pos);
      dprint(1,"dropped on",pos.x(),pos.y(),dropitem and str(dropitem.text(1)));
      # if event originated with ourselves, reorder items
      if ev.source() is self:
        self.reorderItems(dropitem,*files);
      # else event is from someone else, accept the dropped files
      else:
        self._dropped_on = dropitem;
        self.emit(SIGNAL("droppedFiles"),*files);
        # if event originated with another DPTreeWidget, emit a draggedAwayFiles() signal on its behalf
        if isinstance(ev.source(),DPTreeWidget):
          ev.source().emit(SIGNAL("draggedAwayFiles"),*files);

  def reorderItems (self,beforeitem,*files):
    # make list of items to be moved
    moving_items = [];
    for ff in files:
      item = self.dpitems.get(ff,None);
      # if dropping item on top of itself, ignore the operation
      if item is beforeitem:
        return;
      if item:
        dum = QWidget();
        for col in self.ColAction,self.ColRender:
          widget = self.itemWidget(item,self.ColRender);
          widget and widget.setParent(dum);
        moving_items.append(self.takeTopLevelItem(self.indexOfTopLevelItem(item)));
    dprint(1,"moving items",[str(item.text(1)) for item in moving_items]);
    dprint(1,"to before",beforeitem and str(beforeitem.text(1)));
    # move them to specified location
    if moving_items:
      index = self.indexOfTopLevelItem(beforeitem) if beforeitem else self.topLevelItemCount();
      self.insertTopLevelItems(index,moving_items);
      for item in moving_items:
        for col in self.ColAction,self.ColRender:
          self._itemComboBox(item,col);
      self.emit(SIGNAL("updated"));


  def mimeTypes (self):
    return ["text/x-url"];

  def mimeData (self,itemlist):
    mimedata = QMimeData();
    urls = [];
    for item in itemlist:
      if item._dp:
        urls.append(QUrl.fromLocalFile(item._dp.fullpath or item._dp.sourcepath));
    mimedata.setUrls(urls);
    return mimedata;

  def clear (self):
    QTreeWidget.clear(self);
    self.dpitems = {};

  def _itemComboBox (self,item,column):
    """This returns the QComboBox associated with item and column, or creates a new one if
    it hasn't been created yet. The reason we don't create a combobox immediately is because
    the item needs to be inserted into its QTreeWidget first.""";
    if item.treeWidget() is not self:
      return None;
    combobox = self.itemWidget(item,column);
    if not combobox:
      combobox = QComboBox(self);
      # combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents);
      self.setItemWidget(item,column,combobox);
      width = 0;
      for num,(name,icon) in enumerate(item._combobox_option_list[column]):
        if icon:
          combobox.addItem(icon,name);
        else:
          combobox.addItem(name);
        width = max(width,self._fontmetrics.width(name) + (icon and 16 or 0));
      combobox.setCurrentIndex(item._combobox_current_index[column]);
      item._combobox_changed[column] = Kittens.utils.curry(self._updateItemComboBoxIndex,item,column);
      QObject.connect(combobox,SIGNAL("currentIndexChanged(int)"),item._combobox_changed[column]);
      QObject.connect(combobox,SIGNAL("currentIndexChanged(int)"),self._emitUpdatedSignal);
      # resize section if needed
      width += 32;
      if width > self.header().sectionSize(column):
        self.header().resizeSection(column,width);
    return combobox;

  def _emitUpdatedSignal (self,*dum):
    self.emit(SIGNAL("updated"));

  def _updateItemComboBoxIndex (self,item,column,num):
    """Callback for comboboxes: notifies us that a combobox for the given item and column has changed""";
    item._combobox_current_index[column] = num;
    item._combobox_current_value[column] = item._combobox_option_list[column][num][0];

  def setItemPolicy (self,item,policy):
    """Sets the policy of the given item""";
    index = item._combobox_indices[self.ColAction].get(policy,0);
    self._updateItemComboBoxIndex(item,self.ColAction,index);
    combobox = self.itemWidget(item,self.ColAction);
    if combobox:
      combobox.setCurrentIndex(index);

  def getItemDPList (self):
    """Returns list of item,dp pairs corresponding to content of listview.
    Not-yet-saved items will have dp=None.""";
    itemlist = [ (item,item._dp) for item in self.iterator() ];
    return itemlist;

  def _makeDPItem (self,parent,dp,after=None):
    """Creates listview item for data product 'dp', inserts it after item 'after'""";
    if parent:
      item = QTreeWidgetItem(parent,after);
    else:
      item = QTreeWidgetItem();
    item.setTextAlignment(self.ColAction,Qt.AlignRight|Qt.AlignVCenter);
    item.setTextAlignment(self.ColFilename,Qt.AlignRight|Qt.AlignVCenter);
    item.setTextAlignment(self.ColType,Qt.AlignLeft|Qt.AlignVCenter);
    item.setTextAlignment(self.ColRename,Qt.AlignLeft|Qt.AlignVCenter);
    item.setTextAlignment(self.ColRender,Qt.AlignHCenter|Qt.AlignVCenter);
    item.setTextAlignment(self.ColComment,Qt.AlignLeft|Qt.AlignVCenter);
    item.setFlags(Qt.ItemIsSelectable|Qt.ItemIsDragEnabled|Qt.ItemIsDropEnabled|Qt.ItemIsEnabled);
    item._dp = dp;
    # init stuff for combobox functions above
    item._combobox_option_list = {};
    item._combobox_current_index = {};
    item._combobox_current_value = {};
    item._combobox_indices = {};
    item._combobox_changed = {};
    # setup available policies for new or archived items, set initial policy
    if dp.archived:
      item._menu = self._archived_dp_menu;
      item._combobox_option_list[self.ColAction] = policies = self._policy_list_archived;
      policy = "keep";
    else:
      item._combobox_option_list[self.ColAction] = policies = self._policy_list_default;
      policy = dp.policy;
    # create reverse mapping from policy names to indices
    item._combobox_indices[self.ColAction] = dict([ (name,num) for num,(name,icon) in enumerate(policies) ]);
    # init item policy
    self.setItemPolicy(item,policy);
    # set other columns
    basename = os.path.basename(dp.sourcepath);
    name,ext = os.path.splitext(basename);
    item.setText(self.ColFilename,name);
    item.setText(self.ColType,ext);
    item.setToolTip(self.ColFilename,basename);
    item.setToolTip(self.ColType,basename);
    item.setData(self.ColComment,Qt.EditRole,QVariant(dp.comment or ""));
    # make sure new filenames are unique
    filename = _sanitizeFilename(dp.filename);
    if not dp.archived:
      # tack on .tgz prefix onto dirs
      if os.path.isdir(dp.sourcepath) and not filename.endswith(".tgz"):
        filename += ".tgz";
      # form up set of taken names
      taken_names = set();
      for i0,dp0 in self.getItemDPList():
        if dp0.policy not in ["remove","ignore","banish"]:
          taken_names.add(str(i0.text(self.ColRename)));
      # ensure uniqueness of filename
      filename = _makeUniqueFilename(taken_names,filename);
    item.setData(self.ColRename,Qt.EditRole,QVariant(filename));
    # get list of available renderers
    item._renderers = Purr.Render.getRenderers(dp.fullpath or dp.sourcepath);
    item._render = 0;
    item._combobox_option_list[self.ColRender] = [ (name,None) for name in item._renderers ];
    # create reverse mapping from renderer names to indices
    item._combobox_indices[self.ColRender] = dict([ (name,num) for num,name in enumerate(item._renderers) ]);
    # for archived items, try to find renderer in list
    if dp.archived:
      try:
        item._render = item._renderers.index(dp.render);
      except:
        pass;
    self._updateItemComboBoxIndex(item,self.ColRender,item._render);
    # add to map of items
    self.dpitems[dp.fullpath or dp.sourcepath] = item;
    return item;

  def focusOutEvent (self,ev):
    """Redefine focusOut events to stop editing""";
    Kittens.widgets.ClickableTreeWidget.focusOutEvent(self,ev);
    # if focus is going to a child of ours, do nothing
    wid = QApplication.focusWidget();
    while wid:
      if wid is self:
        return;
      wid = wid.parent();
    # else we're truly losing focus -- stop the editor
    self._startOrStopEditing();

  def keyPressEvent (self,ev):
    """Stop editing if enter is pressed""";
    if ev.key() in (Qt.Key_Enter,Qt.Key_Return):
      self._startOrStopEditing();
    elif ev.key() == Qt.Key_Escape:
      self._cancelEditing();
    else:
      Kittens.widgets.ClickableTreeWidget.keyPressEvent(self,ev);

  def _cancelEditing (self):
    dprint(2,"cancelling editor");
    if not self._editing:
      return;
    item0,column0 = self._editing;
    dprint(2,"cancelling editor for",item0.text(1),column0);
    self.closePersistentEditor(*self._editing);
    self._editing = None;
    item0.setText(column0,self._editing_oldtext);

  def _startOrStopEditing (self,item=None,column=None):
    if self._editing:
      if self._editing == (item,column):
        return;
      else:
        item0,column0 = self._editing;
        dprint(2,"closing editor for",item0.text(1),column0);
        self.closePersistentEditor(*self._editing);
        self._editing = None;
        if column0 == self.ColRename:
          item0.setText(self.ColRename,_sanitizeFilename(str(item0.text(self.ColRename))));
    if item and column in [self.ColRename,self.ColComment]:
      self._editing = item,column;
      dprint(2,"opening editor for",item.text(1),column);
      self._editing_oldtext = item.text(column);
      self.setCurrentItem(item,column);
      self.openPersistentEditor(item,column);
      self.itemWidget(item,column).setFocus(Qt.OtherFocusReason);

  def _itemActivated (self,item,col):
    dprint(2,"_itemActivated",item.text(1),col);
    self._startOrStopEditing(item,col);

  def _itemClicked (self,button,item,point,column):
    self._startOrStopEditing(); # stop editing
    if not item or button != Qt.LeftButton:
      return;
    self._startOrStopEditing(item,column);  # start editing if editable column

  def _itemRenamed (self,item,col):
    if col == self.ColRename:
      item.setText(col,_sanitizeFilename(str(item.text(col))));
    self._editing = None;
    self.emit(SIGNAL("updated"));

  def _currentItemChanged(self,item,previous):
    self._startOrStopEditing();

  def _showItemContextMenu (self,item,point,col):
    """Callback for contextMenuRequested() signal. Pops up item menu, if defined""";
    self._startOrStopEditing();
    menu = getattr(item,'_menu',None);
    if menu:
      # self._current_item tells callbacks what item the menu was referring to
      self._current_item = item;
      self.clearSelection();
      self.setItemSelected(item,True);
      menu.exec_(point);
    else:
      self._current_item = None;

  def _copyItemToClipboard (self):
    """Callback for item menu.""";
    dp = self._current_item and getattr(self._current_item,'_dp',None);
    if dp and dp.archived:
      path = dp.fullpath.replace(" ","\\ ");
      QApplication.clipboard().setText(path,QClipboard.Clipboard);
      QApplication.clipboard().setText(path,QClipboard.Selection);

  def _restoreItemFromArchive (self):
    """Callback for item menu.""";
    dp = self._current_item and getattr(self._current_item,'_dp',None);
    if dp and dp.archived:
      dp.restore_from_archive(parent=self);

  def fillDataProducts(self,dps):
    """Fills listview with existing data products""";
    item = None;
    for dp in dps:
      if not dp.ignored:
        item = self._makeDPItem(self,dp,item);
        # ensure combobox widgets are made
        self._itemComboBox(item,self.ColAction);
        self._itemComboBox(item,self.ColRender);

  def addDataProducts(self,dps):
    """Adds new data products to listview. dps is a list of DP objects.
    Returns True if new (non-quiet) DPs are added, or if existing non-quiet dps are updated.
    (this usually tells the main window to wake up)
    """;
    busy = Purr.BusyIndicator();
    wakeup = False;
    # build up list of items to be inserted
    itemlist = [];
    for dp in dps:
      item = self.dpitems.get(dp.sourcepath);
      # If item already exists, it needs to be moved to its new position
      # If takeTopLevelItem() returns None, then item was already removed (this shouldn't happen,
      # but let's be defensive), and we make a new one anyway.
      if item and self.takeTopLevelItem(self.indexOfTopLevelItem(item)):
        itemlist.append(item);
      else:
        itemlist.append(self._makeDPItem(None,dp));
      wakeup = wakeup or not (dp.ignored or dp.quiet);
    # if these DPs were added as a result of a drag-and-drop, we need to insert them in FRONT of the dropped-on item
    if self._dropped_on:
      index = self.indexOfTopLevelItem(self._dropped_on);
    # else insert at end (after=None)
    else:
      index = self.topLevelItemCount();
    if itemlist:
      self.insertTopLevelItems(index,itemlist);
      self.emit(SIGNAL("updated"));
    for item in itemlist:
      # ensure combobox widgets are made
      self._itemComboBox(item,self.ColAction);
      self._itemComboBox(item,self.ColRender);
    return wakeup;

  def dropDataProducts (self,*pathnames):
    """Drops (that is, deletes) new (i.e. non-archived) DP items matching the given pathnames.""";
    trash = QTreeWidget(None);
    updated = False;
    for path in pathnames:
      item = self.dpitems.get(path);
      if item and not item._dp.archived:
        self.takeTopLevelItem(self.indexOfTopLevelItem(item));
        trash.addTopLevelItem(item);
        del self.dpitems[path];
        updated = True;
    if updated:
      self.emit(SIGNAL("updated"));

  def resolveFilenameConflicts (self):
    """Goes through list of DPs to make sure that their destination names
    do not clash. Adjust names as needed. Returns True if some conflicts were resolved.
    """
    taken_names = set();
    resolved = False;
    # iterate through items
    for item,dp in self.getItemDPList():
      # only apply this to saved DPs
      if dp.policy not in ["remove","ignore","banish"]:
        name0 = str(item.text(self.ColRename));
        name = _makeUniqueFilename(taken_names,name0);
        if name != name0:
          item.setText(self.ColRename,name);
          resolved = True;
          self.emit(SIGNAL("updated"));
    return resolved;

  def buildDPList (self):
    """Builds list of data products."""
    updated = False;
    dps = [];
    itemlist = self.getItemDPList();
    # first remove all items marked for removal, in case their names clash with new or renamed items
    for item,dp in itemlist:
      item._policy = item._combobox_current_value[self.ColAction];
      if dp and item._policy == "remove":
        dp.remove_file();
        dp.remove_subproducts();
    # now, make list
    for item,dp in itemlist:
      if item._policy == "remove":
        continue;
      # update renderer and comment
      render  = item._combobox_current_value[self.ColRender];
      comment = str(item.text(self.ColComment));
      if render != dp.render:
        dp.render = render;
        updated = True;
      if comment != dp.comment:
        dp.comment = comment;
        updated = True;
      # archived DPs may need to be renamed, for new ones simply set the name
      if dp.archived:
        updated = updated or dp.rename(str(item.text(self.ColRename)));
      else:
        dp.set_policy(item._policy);
        dp.filename = str(item.text(self.ColRename));
      dps.append(dp);
    return dps,updated;

class LogEntryEditor (QWidget):
  """This class provides a widget for editing log entries.
  """;
  def __init__ (self,parent):
    QWidget.__init__(self,parent);
    # create splitter
    lo = QVBoxLayout(self);
    lo.setMargin(0);
    self.wsplitter = QSplitter(self);
    self.wsplitter.setOrientation(Qt.Vertical);
    self.wsplitter.setChildrenCollapsible(False);
    # self.wsplitter.setFrameStyle(QFrame.Box|QFrame.Raised);
    # self.wsplitter.setMargin(0);
    self.wsplitter.setLineWidth(0);
    lo.addWidget(self.wsplitter);
    # create pane for comment editor
    editorpane = QWidget(self.wsplitter);
    lo_top = QVBoxLayout(editorpane);
    # lo_top.setResizeMode(QLayout.Minimum);
    lo_top.setMargin(0);
    # create comment editor
    # create title and timestamp label, hide timestamp until it is set (below)
    lo_topline = QHBoxLayout(); lo_top.addLayout(lo_topline);
    lo_topline.setMargin(0);
    self.wtoplabel = QLabel("Entry title:",editorpane);
    self.wtimestamp = QLabel("",editorpane);
    lo_topline.addWidget(self.wtoplabel);
    lo_topline.addStretch(1);
    lo_topline.addWidget(self.wtimestamp);
    # add title editor
    self.wtitle   = QLineEdit(editorpane);
    lo_top.addWidget(self.wtitle);
    self.connect(self.wtitle,SIGNAL("textChanged(const QString&)"),self._titleChanged);
    # add comment editor
    # lo_top.addSpacing(5);
    lo_top.addWidget(QLabel("Comments:",editorpane));
    self.wcomment = QTextEdit(editorpane);
    self.wcomment.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    self.comment_doc = QTextDocument(self);
    self.wcomment.setDocument(self.comment_doc);
    self.wcomment.setAcceptRichText(False);
    self.wcomment.setLineWrapMode(QTextEdit.WidgetWidth);
    self.connect(self.comment_doc,SIGNAL("contentsChanged()"),self._commentChanged);
    lo_top.addWidget(self.wcomment);
    # generate frame for the "Data products" listview
    # lo_top.addSpacing(5);
    # create pane for comment editor
    dppane = QWidget(self.wsplitter);
    lo_top = QVBoxLayout(dppane);
    # lo_top.setResizeMode(QLayout.Minimum);
    lo_top.setMargin(0);
    dpline = QWidget(dppane);
    lo_top.addWidget(dpline);
    dpline.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Fixed);
    lo_dpline = QHBoxLayout(dpline);
    lo_dpline.setMargin(0);
    label = QLabel("<nobr>Data products (<u><font color=blue>help</font></u>):</nobr>",dpline)
    label.setToolTip(self.data_product_help);
    lo_dpline.addWidget(label);
    lo_dpline.addStretch(1);
    wnewdp = QPushButton(pixmaps.folder_open.icon(),"Add file...",dpline);
    self.connect(wnewdp,SIGNAL("clicked()"),self._showAddFileDialog);
    wnewdp.setAutoDefault(False);
    wnewdp_dir = QPushButton(pixmaps.folder_open.icon(),"Add dir...",dpline);
    self.connect(wnewdp_dir,SIGNAL("clicked()"),self._showAddDirDialog);
    wnewdp_dir.setAutoDefault(False);
    self._add_dp_dialog = None;
    lo_dpline.addWidget(wnewdp);
    lo_dpline.addWidget(wnewdp_dir);
    # create DP listview
    ndplv = self.wdplv = DPTreeWidget(dppane);
    ndplv.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    lo_top.addWidget(self.wdplv);
    # connect its signals
    QObject.connect(self.wdplv,SIGNAL("updated"),self.setUpdated);
    QObject.connect(self.wdplv,SIGNAL("droppedFiles"),self,SIGNAL("filesSelected"));
    QObject.connect(self.wdplv,SIGNAL("draggedAwayFiles"),self,SIGNAL("draggedAwayFiles"));
    # other internal init
    self.reset();

  data_product_help = \
      """<P><I>Data products</I> are files that will be archived along with this log entry. If "pounce"
      is enabled, PURR will watch your directories for new or updated files, and insert them in this list
      automatically. You can also click on "Add..." to add files by hand.</P>

      <P>Click on the <I>action</I> column to select what to do with a file. Click on the <I>rename</I> column
      to archive the file under a different name. Click on the <I>comment</I> column to enter a comment for the file. Click on the <I>render</I> column to select a rendering for the file. The basic rendering is "link", which makes a simple HTML link to the archived file. Certain file types (e.g. images) may support
      more elaborate renderings.</P>

      <P>The meaning of the different actions is:</P>
      <DL>
      <DT><B>copy</B></DT><DD>data product will be copied to the log.</DD>
      <DT><B>move</B></DT><DD>data product will be moved to the log.</DD>
      <DT><B>ignore</B></DT><DD>data product will not appear in the log, however PURR will keep
      watching it, and will pounce on it if it changes again.</DD>
      <DT><B>banish</B></DT><DD>data product will not appear in the log, and PURR will stop
      watching it (you can un-banish a data product later by adding it via the
      "Add..." button.)</DD>
      <DT><B>keep</B></DT><DD>(for existing products only) retain data product with this log entry.</DD>
      <DT><B>remove</B></DT><DD>(for existing products only) remove data product from this log entry.</DD>
      </DL>
      """;

  def hideEvent (self,event):
    QWidget.hideEvent(self,event);
    if self._add_dp_dialog:
      self._add_dp_dialog.hide();

  def resetDPs (self):
    self.wdplv.clear();

  def reset (self):
    self.wdplv.clear();
    self.entry = self._timestamp = None;
    self.wtimestamp.hide();
    self.setEntryTitle("Untitled entry");
    self.setEntryComment("No comment");
    self._default_dir = ".";
    # _title_changed is True if title was changed or edited manually
    # _comment_changed is True if comment was changed or edited manually
    # _commend_edited is True if comment was eddited since the last time something was
    #   added with addComment()
    self._title_changed = self._comment_changed = self._comment_edited = False;
    # _last_auto_comment is the last comment to have been added via addComment()
    self._last_auto_comment = None;
    self.updated = False;

  def setUpdated (self,updated=True):
    self.updated = updated;
    if updated:
      self.emit(SIGNAL("updated"));

  def setDefaultDirs (self,*dirnames):
    self._default_dirs = dirnames;

  class AddDataProductDialog (QFileDialog):
    """This is a file selection dialog with an extra quick-jump combobox for
    multiple directories""";
    def __init__ (self,parent,directories=False):
      QFileDialog.__init__(self);
      self.setWindowTitle("PURR: Add Data Products");
      self.dirlist = None;
      # add filters
      #self._dirfilter = "Any directories (*)";
      #if hasattr(self,'setNameFilters'):  # qt4.4+
      #  self.setNameFilters(["Any files (*)",self._dirfilter]);
      #else:
      #  self.setFilters(["Any files (*)",self._dirfilter]);
      # connect signals (translates into a SIGNAL with string arguments)
      self.connect(self,SIGNAL("filesSelected(const QStringList&)"),self._filesSelected);
#      self.connect(self,SIGNAL("fileSelected(const QString&)"),self._fileSelected);
      self.connect(self,SIGNAL("currentChanged(const QString&)"),self._fileHighlighted);
      # self.connect(self,SIGNAL("filterSelected(const QString&)"),self._filterSelected);
      # resize selves
      self.config_name = "add-file-dialog";
      width = Config.getint('%s-width'%self.config_name,768);
      height = Config.getint('%s-height'%self.config_name,512);
      self.resize(QSize(width,height));

    def resizeEvent (self,ev):
      QFileDialog.resizeEvent(self,ev);
      sz = ev.size();
      Config.set('%s-width'%self.config_name,sz.width());
      Config.set('%s-height'%self.config_name,sz.height());

    def show (self):
      QFileDialog.show(self);
      self._file = None;

#    def done (self,code):
#      """Workaround for QFileDialog bug: if DirctoryOnly mode, it doesn't actually emit any
#      fileSelected() when OK is pressed. So we catch the selected dir via fileHighlighted()
#      below, and report it here.""";
#      if code == 1 and self.mode() == QFileDialog.DirectoryOnly and self._file:
#        self.emit(SIGNAL("filesSelected"),self._file);
#      QFileDialog.done(self,code);

    def _filterSelected (self,file_filter):
      if str(file_filter) == self._dirfilter:
        self.setFileMode(QFileDialog.Directory);
        self.setOption(QFileDialog.ShowDirsOnly);
      else:
        self.setFileMode(QFileDialog.ExistingFiles);
        self.setOption(0);

    def setDirList (self,dirlist):
      if dirlist is not self.dirlist:
        self.dirlist = dirlist;
        if dirlist and hasattr(self,'setSidebarUrls'):
          self.setSidebarUrls([QUrl.fromLocalFile(path) for path in dirlist]);

    def _filesSelected (self,filelist):
      self.emit(SIGNAL("filesSelected"),*sorted(map(str,filelist)));

    def _fileSelected (self,file):
      self.emit(SIGNAL("filesSelected"),str(file));

    def _fileHighlighted (self,file):
      self._file = str(file);

  def _showAddFileDialog (self,add_dir=False):
    # create dialog when first called
    dialog = self._add_dp_dialog;
    if not dialog:
      self._add_dp_dialog = dialog = self.AddDataProductDialog(self);
      self.connect(dialog,SIGNAL("filesSelected"),self,SIGNAL("filesSelected"));
      dialog.setDirectory(self._default_dirs[0] if self._default_dirs else ".");
    # set mode
    if add_dir:
      dialog.setFileMode(QFileDialog.Directory);
    else:
      dialog.setFileMode(QFileDialog.ExistingFiles);
    # add quick-jump combobox
    dialog.setDirList(list(self._default_dirs)+[os.path.expanduser("~")]);
    dialog.setDirectory(dialog.directory()); # hope this is the same as rereadDir() in qt3
    dialog.show();
    dialog.raise_();

  def _showAddDirDialog (self):
    return self._showAddFileDialog(add_dir=True);

  def _titleChanged (self,*dum):
    self._title_changed = True;
    self.setUpdated();

  def _commentChanged (self,*dum):
    self._comment_changed = self._comment_edited = True;
    self.setUpdated();

  def suggestTitle (self,title):
    """Suggests a title for the entry.
    If title has been manually edited, suggestion is ignored.""";
    if not self._title_changed or not str(self.wtitle.text()):
      self.wtitle.setText(title);
      self._title_changed = False;

  def addComment (self,comment):
    # get current comment text, if nothing was changed at all, use empty text
    if self._comment_changed:
      cur_comment = str(self.comment_doc.toPlainText());
      cpos = self.wcomment.textCursor();
    else:
      cur_comment = "";
      cpos = None;
    # no current comments? Replace
    if not cur_comment:
      cur_comment = comment;
    # else does the comment still end with the last auto-added string? Simply
    # append our comment then
    elif self._last_auto_comment and cur_comment.endswith(self._last_auto_comment):
      cur_comment += comment;
    # else user must've edited the end of the comment. Start a new paragraph and append our
    # comment
    else:
      cur_comment += "\n\n"+comment;
    self._comment_edited = False;
    self._last_auto_comment = comment;
    # update widget
    self.comment_doc.setPlainText(cur_comment);
    if cpos:
      self.wcomment.setTextCursor(cpos);

  def countRemovedDataProducts (self):
    """Returns number of DPs marked for removal""";
    return len([item for item,dp in self.wdplv.getItemDPList() if dp.policy == "remove"]);

  def resolveFilenameConflicts (self,dialog=True):
    """Goes through list of DPs to make sure that their destination names
    do not clash. Applies new names. Returns True if some conflicts were resolved.
    If dialog is True, shows confirrmation dialog."""
    resolved = self.wdplv.resolveFilenameConflicts();
    if resolved and dialog:
      QMessageBox.warning(self,"Filename conflicts","""<P>
        <NOBR>PURR has found duplicate destination filenames among your data products.</NOBR>
        This is not allowed, so some filenames have been adjusted to avoid name clashes.
        Please review the changes before saving this entry.
        </P>""",
        QMessageBox.Ok,0);
    return resolved;

  def updateEntry (self):
    """Updates entry object with current content of dialog.
    In new entry mode (setEntry() not called, so self.entry=None), creates new entry object.
    In old entry mode (setEntry() called), updates and saves old entry object.
    """;
    # form up new entry
    title = str(self.wtitle.text());
    comment = str(self.comment_doc.toPlainText());
    # process comment string -- eliminate single newlines, make double-newlines into separate paragraphs
    # exception are paragraphs that start with "#LOG:", these get special treatment and the single newlines are
    # left intact
    pars = [];
    for paragraph in comment.split("\n\n"):
      if paragraph.startswith("LOG:"):
        pars.append(paragraph.replace("\n","<BR>"));
      else:
        pars.append(paragraph.replace("\n"," "));
    comment = "\n".join(pars);
    # go through data products and decide what to do with each one
    busy = Purr.BusyIndicator();
    # get list of DPs
    dps,updated = self.wdplv.buildDPList();
    # emit signal for all newly-created DPs
    for dp in dps:
      if not dp.archived:
        self.emit(SIGNAL("creatingDataProduct"),dp.sourcepath);
    # update or return new entry
    if self.entry:
      self.entry.update(title=title,comment=comment,dps=dps);
      self.entry.save(refresh_index=updated);
      return self.entry;
    else:
      return Purr.LogEntry(time.time(),title,comment,dps);

  def updateIgnoredEntry (self):
    """Updates an ignore-entry object with current content of dialog, by
    marking all data products for ignore."""
    # collect new DPs from items
    dps = [];
    for item,dp in self.wdplv.getItemDPList():
      if dp and not dp.archived: # None means a new DP
        # set all policies to ignore, unless already set to banish
        if dp.policy != "banish":
          self.wdplv.setItemPolicy(item,"ignore");
          dp.set_policy("ignore");
        dps.append(dp);
    # return new entry
    return Purr.LogEntry(time.time(),dps=dps,ignore=True);

  def setEntry (self,entry=None):
    """Populates the dialog with contents of an existing entry.""";
    busy = Purr.BusyIndicator();
    self.entry = entry;
    self.setEntryTitle(entry.title);
    self.setEntryComment(entry.comment.replace("\n","\n\n").replace("<BR>","\n"));
    self.wdplv.clear();
    self.wdplv.fillDataProducts(entry.dps);
    self.setTimestamp(entry.timestamp);
    self.updated = False;

  def addDataProducts(self,dps):
    """Adds data products to dialog.
      dps is a list of DP objects
    """;
    return self.wdplv.addDataProducts(dps);

  def dropDataProducts (self,*pathnames):
    """Drops new (i.e. non-archived) DP items matching the given pathnames.""";
    return self.wdplv.dropDataProducts(*pathnames);

  def entryTitle (self):
    return self.wtitle.text();

  def setEntryTitle (self,title,select=True):
    self.wtitle.setText(title);
    if select:
      self.wtitle.selectAll();

  def setEntryComment (self,comment,select=True):
    self.comment_doc.setPlainText(comment);
    if select:
      self.wcomment.selectAll();

  def setTimestamp (self,timestamp):
    self._timestamp = timestamp;
    txt = time.strftime("%x %X",time.localtime(timestamp));
    self.wtimestamp.setText(txt);
    self.wtimestamp.show();


class NewLogEntryDialog (QDialog):
  #class DialogTip (QToolTip):
    #def __init__ (self,parent):
      #QToolTip.__init__(self,parent);
      #self.parent = parent;
    #def maybeTip (self,pos):
      #parent = self.parent;
      #if parent._has_tip:
        #rect = QRect(pos.x()-20,pos.y()-20,40,40);
        #self.tip(rect,parent._has_tip);
        #parent._has_tip = None;

  def __init__ (self,parent,*args):
    QDialog.__init__(self,parent,*args);
    self.setWindowTitle("Adding Log Entry");
    self.setWindowIcon(pixmaps.purr_logo.icon());
    self.setModal(False);
    ## create pop-up tip
    #self._has_tip = None;
    #self._dialog_tip = self.DialogTip(self);
    # create editor
    lo = QVBoxLayout(self);
    lo.setMargin(5);
    # lo.setResizeMode(QLayout.Minimum);
    self.editor = LogEntryEditor(self);
    self.dropDataProducts = self.editor.dropDataProducts;
    self.connect(self.editor,SIGNAL("filesSelected"),self,SIGNAL("filesSelected"));
    # connect draggedAwayFiles() signal so that files dragged away are removed from
    # the list
    self.connect(self.editor,SIGNAL("draggedAwayFiles"),self.editor.dropDataProducts);
    lo.addWidget(self.editor);
    lo.addSpacing(5);
    # create button bar
    btnfr = QFrame(self);
    btnfr.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Fixed);
    # btnfr.setMargin(5);
    lo.addWidget(btnfr);
    lo.addSpacing(5);
    btnfr_lo = QHBoxLayout(btnfr);
    btnfr_lo.setMargin(0);
    newbtn = QPushButton(pixmaps.filesave.icon(),"Add new entry",btnfr);
    newbtn.setToolTip("""<P>Saves new log entry, along with any data products not marked
      as "ignore" or "banish".<P>""");
    self.ignorebtn = QPushButton(pixmaps.red_round_cross.icon(),"Ignore all",btnfr);
    self.ignorebtn.setEnabled(False);
    self.ignorebtn.setToolTip("""<P>Tells PURR to ignore all listed data products. PURR
      will not pounce on these files again until they have been modified.<P>""");
    cancelbtn = QPushButton(pixmaps.grey_round_cross.icon(),"Hide",btnfr);
    cancelbtn.setToolTip("""<P>Hides this dialog.<P>""");
    QObject.connect(newbtn,SIGNAL("clicked()"),self.addNewEntry);
    QObject.connect(self.ignorebtn,SIGNAL("clicked()"),self.ignoreAllDataProducts);
    QObject.connect(cancelbtn,SIGNAL("clicked()"),self.hide);
    for btn in (newbtn,self.ignorebtn,cancelbtn):
      btn.setAutoDefault(False);
    btnfr_lo.setMargin(0);
    btnfr_lo.addWidget(newbtn,2);
    btnfr_lo.addStretch(1);
    btnfr_lo.addWidget(self.ignorebtn,2);
    btnfr_lo.addStretch(1);
    btnfr_lo.addWidget(cancelbtn,2);
    self.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding);
    # resize selves
    self.setMinimumSize(256,512);
    width = Config.getint('entry-editor-width',512);
    height = Config.getint('entry-editor-height',512);
    self.resize(QSize(width,height));

  def resizeEvent (self,ev):
    QDialog.resizeEvent(self,ev);
    sz = ev.size();
    Config.set('entry-editor-width',sz.width());
    Config.set('entry-editor-height',sz.height());

  def showEvent (self,ev):
    QDialog.showEvent(self,ev);
    self.emit(SIGNAL("shown"),True);

  def hideEvent (self,ev):
    QDialog.hideEvent(self,ev);
    self.emit(SIGNAL("shown"),False);

  def reset (self):
    self.editor.reset();

  def addDataProducts (self,dps):
    updated = self.editor.addDataProducts(dps);
    self.ignorebtn.setEnabled(True);
    return updated;

  def suggestTitle(self,title):
    self.editor.suggestTitle(title);

  def addComment(self,comment):
    self.editor.addComment(comment);

  def setDefaultDirs (self,*dirs):
    self.editor.setDefaultDirs(*dirs);

  def ignoreAllDataProducts (self):
    # confirm with user
    if QMessageBox.question(self,"Ignoring all data products","""<P><NOBR>Do you really
          want to ignore all data products</NOBR> listed here? PURR will ignore these
          files until they have been modified again.""",
          QMessageBox.Yes,QMessageBox.No) != QMessageBox.Yes:
      return;
    # add ingore entry, which will store info on all ignored data products
    entry = self.editor.updateIgnoredEntry();
    self.emit(SIGNAL("newLogEntry"),entry);
    self.editor.resetDPs();
    self.hide();

  def addNewEntry (self):
    # if some naming conflicts have been resolved, return -- user will need to re-save
    if self.editor.resolveFilenameConflicts():
      return;
    # confirm with user
    if QMessageBox.question(self,"Adding new entry","""<P><NOBR>Do you really want to add
          a new log entry</NOBR> titled "%s"?"""%(self.editor.entryTitle()),
          QMessageBox.Yes,QMessageBox.No) != QMessageBox.Yes:
      return;
    # add entry
    entry = self.editor.updateEntry();
    self.emit(SIGNAL("newLogEntry"),entry);
    self.editor.reset();
    self.hide();

class ExistingLogEntryDialog (QDialog):
  def __init__ (self,parent,*args):
    QDialog.__init__(self,parent,*args);
    self.setModal(False);
    # make stack for viewer and editor components
    lo = QVBoxLayout(self);
    lo.setMargin(5);
    self.wstack = QStackedWidget(self);
    lo.addWidget(self.wstack);
    self.wstack.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    # make editor panel
    self.editor_panel = QWidget(self.wstack);
    self.editor_panel.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    self.wstack.addWidget(self.editor_panel);
    lo = QVBoxLayout(self.editor_panel);
    lo.setMargin(0);
    # create editor
    self.editor = LogEntryEditor(self.editor_panel);
    self.connect(self.editor,SIGNAL("updated"),self._entryUpdated);
    self.connect(self.editor,SIGNAL("filesSelected"),self,SIGNAL("filesSelected"));
    self.connect(self.editor,SIGNAL("creatingDataProduct"),
                 self,SIGNAL("creatingDataProduct"));
    lo.addWidget(self.editor);
    self.dropDataProducts = self.editor.dropDataProducts;
    # create button bar for editor
    btnfr = QFrame(self.editor_panel);
    btnfr.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Fixed);
    # btnfr.setMargin(5);
    lo.addWidget(btnfr);
    lo.addSpacing(5);
    btnfr_lo = QHBoxLayout(btnfr);
    self.wsave = QPushButton(pixmaps.filesave.icon(),"Save",btnfr);
    QObject.connect(self.wsave,SIGNAL("clicked()"),self._saveEntry);
    cancelbtn = QPushButton(pixmaps.grey_round_cross.icon(),"Cancel",btnfr);
    QObject.connect(cancelbtn,SIGNAL("clicked()"),self._cancelEntry);
    for btn in (self.wsave,cancelbtn):
      btn.setAutoDefault(False);
    btnfr_lo.setMargin(0);
    btnfr_lo.addWidget(self.wsave,1);
    btnfr_lo.addStretch(1);
    btnfr_lo.addWidget(cancelbtn,1);

    # create viewer panel
    self.viewer_panel = QWidget(self.wstack);
    self.viewer_panel.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    self.wstack.addWidget(self.viewer_panel);
    lo = QVBoxLayout(self.viewer_panel);
    lo.setMargin(0);
    label = QLabel("""<P>Below is an HTML rendering of your log entry. Note that this is
      only a bare-bones viewer, so only some links will work. In particular,
      you may click on a link associated with a saved data product to get a menu of related actions. To edit this entry, click the Edit button below.
      </P>""",self.viewer_panel);
    label.setWordWrap(True);
    label.setMargin(5);
    lo.addWidget(label);
    self.viewer = self.EntryTextBrowser(self.viewer_panel);
    self.viewer.setOpenLinks(False);
    self.viewer.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    self._viewer_source = None;
    QObject.connect(self.viewer,SIGNAL("anchorClicked(const QUrl &)"),self._urlClicked);
    lo.addWidget(self.viewer);
    lo.addSpacing(5);
    # create button bar
    btnfr = QFrame(self.viewer_panel);
    btnfr.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Fixed);
    # btnfr.setMargin(5);
    lo.addWidget(btnfr);
    lo.addSpacing(5);
    btnfr_lo = QHBoxLayout(btnfr);
    btnfr_lo.setMargin(0);
    btn = self.wprev = QPushButton(pixmaps.previous.icon(),"Previous",btnfr);
    QObject.connect(btn,SIGNAL("clicked()"),self,SIGNAL("previous()"));
    btnfr_lo.addWidget(btn,1);
    btnfr_lo.addSpacing(5);
    btn = self.wnext = QPushButton(pixmaps.next.icon(),"Next",btnfr);
    QObject.connect(btn,SIGNAL("clicked()"),self,SIGNAL("next()"));
    btnfr_lo.addWidget(btn,1);
    btnfr_lo.addSpacing(5);
    btn = self.wedit = QPushButton(pixmaps.edit.icon(),"Edit",btnfr);
    QObject.connect(btn,SIGNAL("clicked()"),self._editEntry);
    btnfr_lo.addWidget(btn,1);
    btnfr_lo.addStretch(1)
    btn = self.wclose = QPushButton(pixmaps.grey_round_cross.icon(),"Close",btnfr);
    QObject.connect(btn,SIGNAL("clicked()"),self.hide);
    btnfr_lo.addWidget(btn,1);
    # create popup menu for data products
    self._dp_menu = menu = QMenu(self);
    self._dp_menu_title = QLabel();
    self._dp_menu_title.setMargin(5);
    self._dp_menu_title_wa = wa=QWidgetAction(self);
    wa.setDefaultWidget(self._dp_menu_title);
    menu.addAction(wa);
    menu.addSeparator();
    menu.addAction(pixmaps.editcopy.icon(),"Restore file(s) from archived copy",self._restoreDPFromArchive);
    menu.addAction(pixmaps.editpaste.icon(),"Copy pathname of archived copy to clipboard",self._copyDPToClipboard);
    menu.addAction("Drag-and-drop archived copy",self._startDPDrag);
    self._dp_menu_on = None;

    # resize selves
    width = Config.getint('entry-viewer-width',512);
    height = Config.getint('entry-viewer-height',512);
    self.resize(QSize(width,height));
    # other init
    self.entry = None;
    self.updated = False;

  class EntryTextBrowser (QTextBrowser):
    def mouseReleaseEvent (self,ev):
      self._mouse_release_pos = ev.globalPos();
      return QTextBrowser.mouseReleaseEvent(self,ev);
    def getLastMouseRelease (self):
      try:
        return self._mouse_release_pos;
      except AttributeError:
        return self.mapToGlobal(QPoint(0,0));

  def resizeEvent (self,ev):
    QDialog.resizeEvent(self,ev);
    sz = ev.size();
    Config.set('entry-viewer-width',sz.width());
    Config.set('entry-viewer-height',sz.height());

  def viewEntry (self,entry,prev=None,next=None):
    # if editing previous entry, ask for confirmation
    if self.updated:
      self.show();
      self._saveEntry();
    busy = Purr.BusyIndicator();
    self.entry = entry;
    self.updated = False;
    self.setWindowTitle(entry.title);
    self._viewer_source = QUrl.fromLocalFile(self.entry.index_file);
    self.viewer.setSource(self._viewer_source);
    self.viewer.reload();
    #self._prev_path = prev and prev.index_file;
    #self._next_path = next and next.index_file;
    self.wprev.setEnabled(bool(prev));
    self.wnext.setEnabled(bool(next));
    self.wstack.setCurrentWidget(self.viewer_panel);

  def _urlClicked (self,url):
    #if self._viewer_source:
    # self.viewer.setSource(self._viewer_source);
    # see if we clicked on a URL for a data product, display URL for it if so
    path = str(url.path());
    if path:
      # see if we clicked on a URL for a data product, display URL for it if so
      for dp in self.entry.dps:
        if os.path.samefile(dp.fullpath,path) or os.path.exists(dp.subproduct_dir()) and os.path.samefile(dp.subproduct_dir(),os.path.dirname(path)):
          self._dp_menu_title.setText(os.path.basename(dp.filename));
          self._dp_menu_on = dp;
          self._dp_menu.exec_(self.viewer.getLastMouseRelease());
          return;
      # else emit signal
      self.emit(SIGNAL("viewPath"),path);


  def _copyDPToClipboard (self):
    """Callback for item menu.""";
    dp = self._dp_menu_on;
    if dp and dp.archived:
      path = dp.fullpath.replace(" ","\\ ");
      QApplication.clipboard().setText(path,QClipboard.Clipboard);
      QApplication.clipboard().setText(path,QClipboard.Selection);

  def _restoreDPFromArchive (self):
    """Callback for item menu.""";
    dp = self._dp_menu_on;
    if dp and dp.archived:
      dp.restore_from_archive(parent=self);

  def _startDPDrag (self):
    """Callback for item menu.""";
    dp = self._dp_menu_on;
    if dp and dp.archived:
      drag = QDrag(self);
      mimedata = QMimeData();
      mimedata.setUrls([QUrl.fromLocalFile(dp.fullpath)]);
      drag.setMimeData(mimedata);
      drag.exec_(Qt.CopyAction|Qt.LinkAction);

  def setDefaultDirs (self,*dirs):
    self.editor.setDefaultDirs(*dirs);

  def addDataProducts (self,dps):
    self.editor.addDataProducts(dps);
    self._entryUpdated();

  def _editEntry (self):
    self.setWindowTitle("Editing entry");
    self.editor.setEntry(self.entry);
    self.updated = False;
    self.wsave.setEnabled(False);
    self.wstack.setCurrentWidget(self.editor_panel);

  def _saveEntry (self):
    # if some naming conflicts have been resolved, return -- user will need to re-save
    if self.editor.resolveFilenameConflicts():
      return;
    # ask for confirmation
    nremove = self.editor.countRemovedDataProducts();
    msg = "<P><nobr>Save changes to this log entry?</nobr></P>";
    if nremove:
      msg += "<P>%d archived data product(s) will be removed.</P>"%nremove;
    if QMessageBox.question(self,"Saving entry",msg,
          QMessageBox.Yes,QMessageBox.No) != QMessageBox.Yes:
      return;
    busy = Purr.BusyIndicator();
    self.editor.updateEntry();
    self.updated = False;
    self.setWindowTitle(self.entry.title);
    self.viewer.reload(); # setSource(QUrl.fromLocalFile(self.entry.index_file));
    self.wstack.setCurrentWidget(self.viewer_panel);
    # emit signal to regenerate log
    self.emit(SIGNAL("entryChanged"),self.entry);

  def _cancelEntry (self):
    if self.updated and QMessageBox.question(self,"Abandoning changes",
          "Abandon changes to this log entry?",
          QMessageBox.Yes,QMessageBox.No) != QMessageBox.Yes:
      return;
    self.setWindowTitle(self.entry.title);
    self.wstack.setCurrentWidget(self.viewer_panel);

  def _entryUpdated (self):
    self.updated = True;
    self.wsave.setEnabled(True);




#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import color
import line_style
import fill_style
import chart_object
import object_set
import pychart_util
import tick_mark_doc
from pychart_types import *

_keys = {
    "line_style": (line_style.T, line_style.default, "The line style of the tick mark."),
    "fill_style": (fill_style.T, fill_style.white, "The fill style."),
    "size": (UnitType, 5, "Size of the tick mark."),
    }

class T(chart_object.T):
    __doc__ = tick_mark_doc.doc
    keys = _keys
##AUTOMATICALLY GENERATED

##END AUTOMATICALLY GENERATED
    
class Circle(T):
    """Draws a circle. """
    def draw(self, can, x, y):
        can.ellipsis(self.line_style, self.fill_style, x, y,
                        self.size/2.0, 1)
        
class Square(T):
    """Draws a square."""
    def draw(self, can, x, y):
        # move to the bottom-left corner
        x = x - self.size/2.0
        y = y - self.size/2.0
        can.rectangle(self.line_style, self.fill_style,
                         x, y, x+self.size, y+self.size)

class Triangle(T):
    """Draws a triangle pointing up."""
    def draw(self, can, x, y):
        can.polygon(self.line_style, self.fill_style,
                       ((x-self.size/1.6, y-self.size/2.0),
                        (x+self.size/1.6, y-self.size/2.0),
                        (x, y+self.size/2.0)))
class DownTriangle(T):
    """Draws a triangle pointing down."""
    def draw(self, can, x, y):
        can.polygon(self.line_style, self.fill_style,
                       ((x, y-self.size/2.0),
                        (x-self.size/1.6, y+self.size/2.0),
                        (x+self.size/1.6, y+self.size/2.0)))


class X(T):
    """Draw a "X"-shaped tick mark. Attribute "fill-style" is ignored."""
    keys = pychart_util.union_dict(T.keys,
                         {"line_style": (line_style.T,
                                         line_style.T(width=0.7),
                                         "The line style of the tick mark")})
    def draw(self, can, x, y):
        # move to the bottom-left corner
        x = x - self.size/2.0
        y = y - self.size/2.0
        can.line(self.line_style, x, y, x+self.size, y+self.size)
        can.line(self.line_style, x+self.size, y, x, y+self.size)
        
class Plus(T):
    """Draw a "+"-shaped tick mark. Attribute "fill-style" is ignored."""
    keys = pychart_util.union_dict(T.keys,
                         {"line_style": (line_style.T, 
                                        line_style.T(width=1),
                                         "The line style of the tick mark.")})
    def draw(self, can, x, y):
        # move to the bottom-left corner
        can.line(self.line_style, x-self.size/1.4, y, x+self.size/1.4, y)
        can.line(self.line_style, x, y-self.size/1.4, x, y+self.size/1.4)
        
class Diamond(T):
    """Draw a square rotated at 45 degrees."""
    def draw(self, can, x, y):
        # move to the bottom-left corner
        can.polygon(self.line_style, self.fill_style,
                   ((x-self.size/1.4, y), (x, y+self.size/1.4),
                    (x+self.size/1.4, y), (x, y-self.size/1.4)))

class Star(T):
    """Draw a "*". Attribute "fill-style" is ignored."""
    keys = pychart_util.union_dict(T.keys,
                         {"line_style": (line_style.T,
                                        line_style.T(width=1),
                                         "The line style of the tick mark.")})
    def draw(self, can, x, y):
        # move to the bottom-left corner
        midx = x
        midy = y
        d_len = self.size / 2.0
        r_len = self.size * 1.414 / 2.0
        can.line(self.line_style, x-d_len, y-d_len, x+d_len, y+d_len)
        can.line(self.line_style, x+d_len, y-d_len, x-d_len, y+d_len) 
        can.line(self.line_style, midx, y-r_len, midx, y+r_len)
        can.line(self.line_style, x-r_len, midy, x+r_len, midy)
        
class Null(T):
    """This tickmark doesn't draw anything. All the attributes are ignored."""
    def __init__ (self):
        self.line_style = None
        self.fill_style = None
        self.size = -1
    def draw(self, can, x, y):
        pass
    
standards = object_set.T()
def _intern(style):
    standards.add(style)
    return style
     
square = _intern(Square())
square3 = _intern(Square(size=3))
square5 = square
x = _intern(X())
x3 = _intern(X(size=3))
x5 = x
star = _intern(Star())
star3 = _intern(Star(size=3))
star5 = star
plus = _intern(Plus())
plus3 = _intern(Plus(size=3))
plus5 = plus
dia = _intern(Diamond())
dia3 = _intern(Diamond(size=3))
dia5 = dia
tri = _intern(Triangle())
tri3 = _intern(Triangle(size=3))
tri5 = tri
dtri = _intern(DownTriangle())
dtri3 = _intern(DownTriangle(size=3))
dtri5 = dtri
circle1 = _intern(Circle(size=1))
circle2 = _intern(Circle(size=3))
circle3 = _intern(Circle(size=5))
blacksquare = _intern(Square(fill_style=fill_style.black))
blacksquare3 = _intern(Square(size=3, fill_style=fill_style.black))
blackdia = _intern(Diamond(fill_style=fill_style.black))
blackdia3 = _intern(Diamond(size=3, fill_style=fill_style.black))
blacktri = _intern(Triangle(fill_style=fill_style.black))
blacktri3 = _intern(Triangle(size=3, fill_style=fill_style.black))
blackdtri = _intern(DownTriangle(fill_style=fill_style.black))
blackdtri3 = _intern(DownTriangle(size=3, fill_style=fill_style.black))
blackcircle1 = _intern(Circle(size=1, fill_style=fill_style.black))
blackcircle3 = _intern(Circle(size=3, fill_style=fill_style.black))
gray70square = _intern(Square(fill_style=fill_style.gray70))
gray70square3 = _intern(Square(size=3, fill_style=fill_style.gray70))
gray70dia = _intern(Diamond(fill_style=fill_style.gray70))
gray70dia3 = _intern(Diamond(size=3, fill_style=fill_style.gray70))
gray70tri = _intern(Triangle(fill_style=fill_style.gray70))
gray70tri3 = _intern(Triangle(size=3, fill_style=fill_style.gray70))
gray70dtri = _intern(DownTriangle(fill_style=fill_style.gray70))
gray70dtri3 = _intern(DownTriangle(size=3, fill_style=fill_style.gray70))
gray70circle1 = _intern(Circle(size=1, fill_style=fill_style.gray70))
gray70circle3 = _intern(Circle(size=3, fill_style=fill_style.gray70))
default = _intern(Null())




# -*- coding: utf-8 -*-
_tdl_no_reimport = True;

import os.path
import time

from PyQt4.Qt import *

import Purr
import Purr.Editors
import Purr.LogEntry
import Purr.Pipe
import Purr.RenderIndex
from Purr import Config,pixmaps,dprint,dprintf
import Kittens.widgets
import Kittens.utils

class BusyIndicator (object):
  def __init__ (self):
    QApplication.setOverrideCursor(QCursor(Qt.WaitCursor));
  def __del__ (self):
    QApplication.restoreOverrideCursor();

class HTMLViewerDialog (QDialog):
  """This class implements a dialog to view a piece of HTML text.""";
  def __init__ (self,parent,config_name=None,buttons=[],*args):
    """Creates dialog.
    'config_name' is used to get/set default window size from Config object
    'buttons' can be a list of names or (QPixmapWrapper,name[,tooltip]) tuples to provide
    custom buttons at the bottom of the dialog. When a button is clicked, the dialog
    emits SIGNAL("name").
    A "Close" button is always provided, this simply hides the dialog.
    """;
    QDialog.__init__(self,parent,*args);
    self.setModal(False);
    lo = QVBoxLayout(self);
    # create viewer
    self.label = QLabel(self);
    self.label.setMargin(5);
    self.label.setWordWrap(True);
    lo.addWidget(self.label);
    self.label.hide();
    self.viewer = QTextBrowser(self);
    lo.addWidget(self.viewer);
    # self.viewer.setReadOnly(True);
    self.viewer.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    QObject.connect(self.viewer,SIGNAL("anchorClicked(const QUrl &)"),self._urlClicked);
    self._source = None;
    lo.addSpacing(5);
    # create button bar
    btnfr = QFrame(self);
    btnfr.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Fixed);
    # btnfr.setMargin(5);
    lo.addWidget(btnfr);
    lo.addSpacing(5);
    btnfr_lo = QHBoxLayout(btnfr);
    btnfr_lo.setMargin(5);
    # add user buttons
    self._user_buttons = {};
    for name in buttons:
      if isinstance(name,str):
        btn = QPushButton(name,btnfr);
      elif isinstance(name,(list,tuple)):
        if len(name) < 3:
          pixmap,name = name;
          tip = None;
        else:
          pixmap,name,tip = name;
        btn = QPushButton(pixmap.icon(),name,btnfr);
        if tip:
          btn.setToolTip(tip);
      self._user_buttons[name] = btn;
      btn._clicked = Kittens.utils.curry(self.emit,SIGNAL(name));
      self.connect(btn,SIGNAL("clicked()"),btn._clicked);
      btnfr_lo.addWidget(btn,1);
    # add a Close button
    btnfr_lo.addStretch(100);
    closebtn = QPushButton(pixmaps.grey_round_cross.icon(),"Close",btnfr);
    self.connect(closebtn,SIGNAL("clicked()"),self.hide);
    btnfr_lo.addWidget(closebtn,1);
    # resize selves
    self.config_name = config_name or "html-viewer";
    width = Config.getint('%s-width'%self.config_name,512);
    height = Config.getint('%s-height'%self.config_name,512);
    self.resize(QSize(width,height));

  def resizeEvent (self,ev):
    QDialog.resizeEvent(self,ev);
    sz = ev.size();
    Config.set('%s-width'%self.config_name,sz.width());
    Config.set('%s-height'%self.config_name,sz.height());

  def setDocument (self,filename,empty=""):
    """Sets the HTML text to be displayed. """;
    self._source = QUrl.fromLocalFile(filename);
    if os.path.exists(filename):
      self.viewer.setSource(self._source);
    else:
      self.viewer.setText(empty);

  def _urlClicked (self,url):
    path = str(url.path());
    if path:
      self.emit(SIGNAL("viewPath"),path);
    # to make sure it keeps displaying the same thing
    self.viewer.setSource(self._source);

  def reload (self):
    self.viewer.reload();

  def setLabel (self,label=None):
    if label is None:
      self.label.hide();
    else:
      self.label.setText(label);
      self.label.show();



      #self.watching = bool(enable);
      #if enable:
      #  self.purrer.enableWatching(self.path);
      #else:
      #  self.purrer.disableWatching(self.path);
      #self.mainwin._checkPounceStatus();



class DirectoryListWidget (Kittens.widgets.ClickableListWidget):
  """This class implements a QTreeWidget for data products.
  """;
  def __init__ (self,*args):
    Kittens.widgets.ClickableListWidget.__init__(self,*args);
    # insert columns, and numbers for them
    # setup other properties of the listview
    self.setAcceptDrops(True);
    self.setSelectionMode(QListWidget.SingleSelection);
    self.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Minimum);
    QObject.connect(self,SIGNAL("itemSelectionChanged()"),self._selectionChanged);
    QObject.connect(self,SIGNAL("itemChanged(QListWidgetItem *)"),self._itemChanged);
    self._items = {};
    self._item_height = None;
    self._max_height_items = 5;
    self._min_height = 8;
    self.setMaximumSize(1000000,self._min_height);

  class DirItem (QListWidgetItem):
      # these represent the watch-states
      ToCheckState =  { Purr.UNWATCHED:Qt.Unchecked,Purr.WATCHED:Qt.PartiallyChecked,Purr.POUNCE:Qt.Checked };
      FromCheckState = dict([(state,watch) for watch,state in ToCheckState.iteritems()]);

      def __init__ (self,pathname,parent=None,watching=Purr.WATCHED):
        self._pathname = pathname;
        pathname = Kittens.utils.collapseuser(pathname);
        self._in_setWatching = True;
        QListWidgetItem.__init__(self,pathname,parent);
        self.setFlags(Qt.ItemIsSelectable|Qt.ItemIsUserCheckable|Qt.ItemIsEnabled|Qt.ItemIsTristate);
        self.setWatching(watching);

      def pathname (self):
        return self._pathname;

      def watching (self):
        return self.FromCheckState[self.checkState()];

      def setWatching (self,watching):
        self._in_setWatching = True;
        self._watching = watching;
        self.setCheckState(self.ToCheckState[watching]);
        self._in_setWatching = False;

  def _selectionChanged (self):
    sel = self.selectedItems();
    self.emit(SIGNAL("directorySelected"),sel and sel[0].pathname());
    self.emit(SIGNAL("hasSelection"),bool(sel));

  def _itemChanged (self,item):
    if not getattr(item,'_in_setWatching',False):
      item.setWatching((item._watching+1)%(Purr.POUNCE+1));
      self.emit(SIGNAL("directoryStateChanged"),item.pathname(),item.watching());

  def _checkSize (self):
    """Automatically resizes widget to display at most max_height_items items""";
    if self._item_height is not None:
      sz = min(self._max_height_items,self.count())*self._item_height + 5;
      sz= max(sz,20);
      self.setMinimumSize(0,sz);
      self.setMaximumSize(1000000,sz);
      self.resize(self.width(),sz);

  def clear (self):
    Kittens.widgets.ClickableListWidget.clear(self);
    self._items = {};

  def add (self,pathname,watching=Purr.WATCHED):
    if pathname in self._items:
      return;
    item = self._items[pathname] = self.DirItem(pathname,parent=self,watching=watching);
    # init item height, if inserting first item
    if self._item_height is None:
      self._item_height = self.visualItemRect(item).height();
    self._checkSize();

  def remove (self,pathname):
    item = self._items.get(pathname,None);
    return item and self._removeItem(item);

  def _removeItem (self,item):
    del self._items[item.pathname()];
    self.emit(SIGNAL("directoryStateChanged"),item.pathname(),Purr.REMOVED);
    self.takeItem(self.row(item));
    self._checkSize();

  def removeCurrent (self,confirm=True):
    sel = self.selectedItems();
    if sel:
      if confirm:
        if QMessageBox.warning(self,"Detaching from directory","""Do you really want to stop monitoring the directory <tt>%s</tt>?"""%sel[0].pathname(),
              QMessageBox.Yes,QMessageBox.No) == QMessageBox.Yes:
          return self._removeItem(sel[0]);
    return None;

class LogEntryTree (Kittens.widgets.ClickableTreeWidget):
  def __init__ (self,*args):
    Kittens.widgets.ClickableTreeWidget.__init__(self,*args);
    self.setDragEnabled(True);

  def mimeTypes (self):
    return ["text/x-url"];

  def mimeData (self,itemlist):
    mimedata = QMimeData();
    urls = [];
    for item in itemlist:
      dp = getattr(item,"_dp",None);
      dp and urls.append(QUrl.fromLocalFile(dp.fullpath or dp.sourcepath));
    mimedata.setUrls(urls);
    return mimedata;


class MainWindow (QMainWindow):

  about_message = """
    <P>PURR ("<B>P</B>URR is <B>U</B>seful for <B>R</B>emembering <B>R</B>eductions", for those working with
    a stable version, or "<B>P</B>URR <B>U</B>sually <B>R</B>emembers <B>R</B>eductions", for those
    working with a development version, or "<B>P</B>URR <B>U</B>sed to <B>R</B>emember <B>R</B>eductions",
    for those working with a broken version) is a tool for
    automatically keeping a log of your data reduction operations. PURR will monitor your working directories
    for new or updated files (called "data products"), and upon seeing any, it can "pounce" -- that is, offer
    you the option of saving the files to a log, along with descriptive comments. It will then
    generate an HTML page with a pretty rendering of your log and data products.</P>
  """;

  def __init__ (self,parent,hide_on_close=False):
    QMainWindow.__init__(self,parent);
    self._hide_on_close = hide_on_close;
    # replace the BusyIndicator class with a GUI-aware one
    Purr.BusyIndicator = BusyIndicator;
    self._pounce = False;
    # we keep a small stack of previously active purrers. This makes directory changes
    # faster (when going back and forth between dirs)
    # current purrer
    self.purrer = None;
    self.purrer_stack = [];
    # Purr pipes for receiving remote commands
    self.purrpipes = {}
    # init GUI
    self.setWindowTitle("PURR");
    self.setWindowIcon(pixmaps.purr_logo.icon());
    cw = QWidget(self);
    self.setCentralWidget(cw);
    cwlo = QVBoxLayout(cw);
    cwlo.setContentsMargins(0,0,0,0);
    cwlo.setMargin(5);
    cwlo.setSpacing(0);
    toplo = QHBoxLayout(); cwlo.addLayout(toplo);

    # About dialog
    self._about_dialog = QMessageBox(self);
    self._about_dialog.setWindowTitle("About PURR");
    self._about_dialog.setText(self.about_message + """
        <P>PURR is not watching any directories right now. You may need to restart it, and give it
  some directory names on the command line.</P>""");
    self._about_dialog.setIconPixmap(pixmaps.purr_logo.pm());
    # Log viewer dialog
    self.viewer_dialog = HTMLViewerDialog(self,config_name="log-viewer",
          buttons=[(pixmaps.blue_round_reload,"Regenerate",
                   """<P>Regenerates your log's HTML code from scratch. This can be useful if
                   your PURR version has changed, or if there was an error of some kind
                   the last time the files were generated.</P>
                   """)]);
    self._viewer_timestamp = None;
    self.connect(self.viewer_dialog,SIGNAL("Regenerate"),self._regenerateLog);
    self.connect(self.viewer_dialog,SIGNAL("viewPath"),self._viewPath);

    # Log title toolbar
    title_tb = QToolBar(cw);
    title_tb.setToolButtonStyle(Qt.ToolButtonTextBesideIcon);
    title_tb.setIconSize(QSize(16,16));
    cwlo.addWidget(title_tb);
    title_label = QLabel("Purrlog title:",title_tb);
    title_tb.addWidget(title_label);
    self.title_editor = QLineEdit(title_tb);
    title_tb.addWidget(self.title_editor);
    self.connect(self.title_editor,SIGNAL("editingFinished()"),self._titleChanged);
    tip = """<P>This is your current log title. To rename the log, enter new name here and press Enter.</P>"""
    title_label.setToolTip(tip);
    self.title_editor.setToolTip(tip);
    self.wviewlog = title_tb.addAction(pixmaps.openbook.icon(),"View",self._showViewerDialog);
    self.wviewlog.setToolTip("Click to see an HTML rendering of your current log.");
    qa = title_tb.addAction(pixmaps.purr_logo.icon(),"About...",self._about_dialog.exec_);
    qa.setToolTip("<P>Click to see the About... dialog, which will tell you something about PURR.</P>");

    self.wdirframe = QFrame(cw);
    cwlo.addWidget(self.wdirframe);
    self.dirs_lo = QVBoxLayout(self.wdirframe);
    self.dirs_lo.setMargin(5);
    self.dirs_lo.setContentsMargins(5,0,5,5);
    self.dirs_lo.setSpacing(0);
    self.wdirframe.setFrameStyle(QFrame.Box|QFrame.Raised);
    self.wdirframe.setLineWidth(1);

    ## Directories toolbar
    dirs_tb = QToolBar(self.wdirframe);
    dirs_tb.setToolButtonStyle(Qt.ToolButtonIconOnly);
    dirs_tb.setIconSize(QSize(16,16));
    self.dirs_lo.addWidget(dirs_tb);
    label = QLabel("Monitoring directories:",dirs_tb);
    self._dirs_tip = """<P>PURR can monitor your working directories for new or updated files. If there's a checkmark
      next to the directory name in this list, PURR is monitoring it.</P>

      <P>If the checkmark is grey, PURR is monitoring things unobtrusively. When a new or updated file is detected in he monitored directory,
      it is quietly added to the list of files in the "New entry" window, even if this window is not currently visible.</P>

      <P>If the checkmark is black, PURR will be more obtrusive. Whenever a new or updated file is detected, the "New entry" window will
      pop up automatically. This is called "pouncing", and some people find it annoying.</P>
      """
    label.setToolTip(self._dirs_tip);
    label.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Minimum);
    dirs_tb.addWidget(label);

    # add directory list widget
    self.wdirlist = DirectoryListWidget(self.wdirframe);
    self.wdirlist.setToolTip(self._dirs_tip);
    QObject.connect(self.wdirlist,SIGNAL("directoryStateChanged"),self._changeWatchedDirState);
    self.dirs_lo.addWidget(self.wdirlist);
    # self.wdirlist.setMaximumSize(1000000,64)

    # add directory button
    add = dirs_tb.addAction(pixmaps.list_add.icon(),"Add",self._showAddDirectoryDialog);
    add.setToolTip("<P>Click to add another directory to be monitored.</P>");

    # remove directory button
    delbtn = dirs_tb.addAction(pixmaps.list_remove.icon(),"Remove",self.wdirlist.removeCurrent);
    delbtn.setEnabled(False);
    delbtn.setToolTip("<P>Click to removed the currently selected directory from the list.</P>");
    QObject.connect(self.wdirlist,SIGNAL("hasSelection"),delbtn.setEnabled);

#    # qa = dirs_tb.addAction(pixmaps.blue_round_reload.icon(),"Rescan",self._forceRescan);
#    # qa.setToolTip("Click to rescan the directories for any new or updated files.");
#    self.wshownew = QCheckBox("show new files",dirs_tb);
#    dirs_tb.addWidget(self.wshownew);
#    self.wshownew.setCheckState(Qt.Checked);
#    self.wshownew.setToolTip("""<P>If this is checked, the "New entry" window will pop up automatically whenever
#  new or updated files are detected. If this is unchecked, the files will be added to the window quietly
#        and unobtrusively; you can show the window manually by clicking on the "New entry..." button below.</P>""");
#    self._dir_entries = {};

    cwlo.addSpacing(5);

    wlogframe = QFrame(cw);
    cwlo.addWidget(wlogframe);
    log_lo = QVBoxLayout(wlogframe);
    log_lo.setMargin(5);
    log_lo.setContentsMargins(5,5,5,5);
    log_lo.setSpacing(0);
    wlogframe.setFrameStyle(QFrame.Box|QFrame.Raised);
    wlogframe.setLineWidth(1);

# listview of log entries
    self.etw = LogEntryTree(cw);
    log_lo.addWidget(self.etw,1);
    self.etw.header().setDefaultSectionSize(128);
    self.etw.header().setMovable(False);
    self.etw.setHeaderLabels(["date","entry title","comment"]);
    if hasattr(QHeaderView,'ResizeToContents'):
      self.etw.header().setResizeMode(0,QHeaderView.ResizeToContents);
    else:
      self.etw.header().setResizeMode(0,QHeaderView.Custom);
      self.etw.header().resizeSection(0,120);
    self.etw.header().setResizeMode(1,QHeaderView.Interactive);
    self.etw.header().setResizeMode(2,QHeaderView.Stretch);
    self.etw.header().show();
    try: self.etw.setAllColumnsShowFocus(True);
    except AttributeError: pass; # Qt 4.2+
    # self.etw.setShowToolTips(True);
    self.etw.setSortingEnabled(False);
    # self.etw.setColumnAlignment(2,Qt.AlignLeft|Qt.AlignTop);
    self.etw.setSelectionMode(QTreeWidget.ExtendedSelection);
    self.etw.setRootIsDecorated(True);
    self.connect(self.etw,SIGNAL("itemSelectionChanged()"),self._entrySelectionChanged);
    self.connect(self.etw,SIGNAL("itemActivated(QTreeWidgetItem*,int)"),self._viewEntryItem);
    self.connect(self.etw,SIGNAL("itemContextMenuRequested"),self._showItemContextMenu);
    # create popup menu for data products
    self._archived_dp_menu = menu = QMenu(self);
    self._archived_dp_menu_title = QLabel();
    self._archived_dp_menu_title.setMargin(5);
    self._archived_dp_menu_title_wa = wa=QWidgetAction(self);
    wa.setDefaultWidget(self._archived_dp_menu_title);
    menu.addAction(wa);
    menu.addSeparator();
    menu.addAction(pixmaps.editcopy.icon(),"Restore file(s) from archived copy",self._restoreItemFromArchive);
    menu.addAction(pixmaps.editpaste.icon(),"Copy pathname of archived copy to clipboard",self._copyItemToClipboard);
    self._current_item = None;
    # create popup menu for entries
    self._entry_menu = menu = QMenu(self);
    self._entry_menu_title = QLabel();
    self._entry_menu_title.setMargin(5);
    self._entry_menu_title_wa = wa=QWidgetAction(self);
    wa.setDefaultWidget(self._entry_menu_title);
    menu.addAction(wa);
    menu.addSeparator();
    menu.addAction(pixmaps.filefind.icon(),"View this log entry",self._viewEntryItem);
    menu.addAction(pixmaps.editdelete.icon(),"Delete this log entry",self._deleteSelectedEntries);
    # buttons at bottom
    log_lo.addSpacing(5);
    btnlo = QHBoxLayout();
    log_lo.addLayout(btnlo);
    self.wnewbtn = QPushButton(pixmaps.filenew.icon(),"New entry...",cw);
    self.wnewbtn.setToolTip("Click to add a new log entry.");
    # self.wnewbtn.setFlat(True);
    self.wnewbtn.setEnabled(False);
    btnlo.addWidget(self.wnewbtn);
    btnlo.addSpacing(5);
    self.weditbtn = QPushButton(pixmaps.filefind.icon(),"View entry...",cw);
    self.weditbtn.setToolTip("Click to view or edit the selected log entry/");
    # self.weditbtn.setFlat(True);
    self.weditbtn.setEnabled(False);
    self.connect(self.weditbtn,SIGNAL("clicked()"),self._viewEntryItem);
    btnlo.addWidget(self.weditbtn);
    btnlo.addSpacing(5);
    self.wdelbtn = QPushButton(pixmaps.editdelete.icon(),"Delete",cw);
    self.wdelbtn.setToolTip("Click to delete the selected log entry or entries.");
    # self.wdelbtn.setFlat(True);
    self.wdelbtn.setEnabled(False);
    self.connect(self.wdelbtn,SIGNAL("clicked()"),self._deleteSelectedEntries);
    btnlo.addWidget(self.wdelbtn);
    # enable status line
    self.statusBar().show();
    Purr.progressMessage = self.message;
    self._prev_msg = None;
    # editor dialog for new entry
    self.new_entry_dialog = Purr.Editors.NewLogEntryDialog(self);
    self.connect(self.new_entry_dialog,SIGNAL("newLogEntry"),self._newLogEntry);
    self.connect(self.new_entry_dialog,SIGNAL("filesSelected"),self._addDPFiles);
    self.connect(self.wnewbtn,SIGNAL("clicked()"),self.new_entry_dialog.show);
    self.connect(self.new_entry_dialog,SIGNAL("shown"),self._checkPounceStatus);
    # entry viewer dialog
    self.view_entry_dialog = Purr.Editors.ExistingLogEntryDialog(self);
    self.connect(self.view_entry_dialog,SIGNAL("previous()"),self._viewPrevEntry);
    self.connect(self.view_entry_dialog,SIGNAL("next()"),self._viewNextEntry);
    self.connect(self.view_entry_dialog,SIGNAL("viewPath"),self._viewPath);
    self.connect(self.view_entry_dialog,SIGNAL("filesSelected"),self._addDPFilesToOldEntry);
    self.connect(self.view_entry_dialog,SIGNAL("entryChanged"),self._entryChanged);
    # saving a data product to an older entry will automatically drop it from the
    # new entry dialog
    self.connect(self.view_entry_dialog,SIGNAL("creatingDataProduct"),
                  self.new_entry_dialog.dropDataProducts);
    # resize selves
    width = Config.getint('main-window-width',512);
    height = Config.getint('main-window-height',512);
    self.resize(QSize(width,height));
    # create timer for pouncing
    self._timer = QTimer(self);
    self.connect(self._timer,SIGNAL("timeout()"),self._rescan);
    # create dict mapping index.html paths to entry numbers
    self._index_paths = {};


  def resizeEvent (self,ev):
    QMainWindow.resizeEvent(self,ev);
    sz = ev.size();
    Config.set('main-window-width',sz.width());
    Config.set('main-window-height',sz.height());

  def closeEvent (self,ev):
    if self._hide_on_close:
      ev.ignore();
      self.hide();
      self.new_entry_dialog.hide();
    else:
      if self.purrer:
        self.purrer.detach();
      return QMainWindow.closeEvent(self,ev);

  def message (self,msg,ms=2000,sub=False):
    if sub:
      if self._prev_msg:
        msg = ": ".join((self._prev_msg,msg));
    else:
      self._prev_msg = msg;
    self.statusBar().showMessage(msg,ms);
    QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents);

  def _changeWatchedDirState (self,pathname,watching):
    self.purrer.setWatchingState(pathname,watching);
    # update dialogs if dir list has changed
    if watching == Purr.REMOVED:
      self.purrpipes.pop(pathname);
      dirs = [ path for path,state in self.purrer.watchedDirectories() ];
      self.new_entry_dialog.setDefaultDirs(*dirs);
      self.view_entry_dialog.setDefaultDirs(*dirs);
    pass;

  def _showAddDirectoryDialog (self):
    dd = str(QFileDialog.getExistingDirectory(self,"PURR: Add a directory to monitor")).strip();
    if dd:
      # adds a watched directory. Default initial setting of 'watching' is POUNCE if all
      # directories are in POUNCE state, or WATCHED otherwise.
      watching = max(Purr.WATCHED,min([state for path,state in self.purrer.watchedDirectories()] or [Purr.WATCHED]));
      self.purrer.addWatchedDirectory(dd,watching);
      self.purrpipes[dd] = Purr.Pipe.open(dd);
      self.wdirlist.add(dd,watching);
      # update dialogs since dir list has changed
      dirs = [ path for path,state in self.purrer.watchedDirectories() ];
      self.new_entry_dialog.setDefaultDirs(*dirs);
      self.view_entry_dialog.setDefaultDirs(*dirs);

  def detachPurrlog (self):
    self.wdirlist.clear();
    self.purrer and self.purrer.detach();
    self.purrer = None;

  def hasPurrlog (self):
    return bool(self.purrer);

  def attachPurrlog (self,purrlog,watchdirs=[]):
    """Attaches Purr to the given purrlog directory. Arguments are passed to Purrer object as is."""
    # check purrer stack for a Purrer already watching this directory
    dprint(1,"attaching to purrlog",purrlog);
    for i,purrer in enumerate(self.purrer_stack):
      if os.path.samefile(purrer.logdir,purrlog):
        dprint(1,"Purrer object found on stack (#%d),reusing\n",i);
        # found? move to front of stack
        self.purrer_stack.pop(i);
        self.purrer_stack.insert(0,purrer);
        # update purrer with watched directories, in case they have changed
        for dd in (watchdirs or []):
          purrer.addWatchedDirectory(dd,watching=None);
        break;
    # no purrer found, make a new one
    else:
      dprint(1,"creating new Purrer object");
      try:
        purrer = Purr.Purrer(purrlog,watchdirs);
      except Purr.Purrer.LockedError,err:
        # check that we could attach, display message if not
        QMessageBox.warning(self,"Catfight!","""<P><NOBR>It appears that another PURR process (%s)</NOBR>
          is already attached to <tt>%s</tt>, so we're not allowed to touch it. You should exit the other PURR
          process first.</P>"""%(err.args[0],os.path.abspath(purrlog)),QMessageBox.Ok,0);
        return False;
      except Purr.Purrer.LockFailError,err:
        QMessageBox.warning(self,"Failed to obtain lock","""<P><NOBR>PURR was unable to obtain a lock</NOBR>
          on directory <tt>%s</tt> (error was "%s"). The most likely cause is insufficient permissions.</P>"""%(os.path.abspath(purrlog),err.args[0]),QMessageBox.Ok,0);
        return False;
      self.purrer_stack.insert(0,purrer);
      # discard end of stack
      self.purrer_stack = self.purrer_stack[:3];
      # attach signals
      self.connect(purrer,SIGNAL("disappearedFile"),
                   self.new_entry_dialog.dropDataProducts);
      self.connect(purrer,SIGNAL("disappearedFile"),
                   self.view_entry_dialog.dropDataProducts);
    # have we changed the current purrer? Update our state then
    # reopen Purr pipes
    self.purrpipes = {};
    for dd,state in purrer.watchedDirectories():
      self.purrpipes[dd] = Purr.Pipe.open(dd);
    if purrer is not self.purrer:
      self.message("Attached to %s"%purrer.logdir,ms=10000);
      dprint(1,"current Purrer changed, updating state");
      # set window title
      path = Kittens.utils.collapseuser(os.path.join(purrer.logdir,''));
      self.setWindowTitle("PURR - %s"%path);
      # other init
      self.purrer = purrer;
      self.new_entry_dialog.hide();
      self.new_entry_dialog.reset();
      dirs = [ path for path,state in purrer.watchedDirectories() ];
      self.new_entry_dialog.setDefaultDirs(*dirs);
      self.view_entry_dialog.setDefaultDirs(*dirs);
      self.view_entry_dialog.hide();
      self.viewer_dialog.hide();
      self._viewing_ientry = None;
      self._setEntries(self.purrer.getLogEntries());
#      print self._index_paths;
      self._viewer_timestamp = None;
      self._updateViewer();
      self._updateNames();
      # update directory widgets
      self.wdirlist.clear();
      for pathname,state in purrer.watchedDirectories():
        self.wdirlist.add(pathname,state);
      # Reset _pounce to false -- this will cause checkPounceStatus() into a rescan
      self._pounce = False;
      self._checkPounceStatus();
    return True;

  def setLogTitle (self,title):
    if self.purrer:
      if title != self.purrer.logtitle:
        self.purrer.setLogTitle(title);
        self._updateViewer();
      self._updateNames();

  def _updateNames (self):
    self.wnewbtn.setEnabled(True);
    self.wviewlog.setEnabled(True);
    self._about_dialog.setText(self.about_message + """
      <P>Your current log resides in:<PRE>  <tt>%s</tt></PRE>To see your log in all its HTML-rendered
      glory, point your browser to <tt>index.html</tt> therein, or use the handy "View" button provided by PURR.</P>

      <P>Your current working directories are:</P>
      <P>%s</P>
      """%(self.purrer.logdir,
      "".join([ "<PRE>  <tt>%s</tt></PRE>"%name for name,state in self.purrer.watchedDirectories()])
    ));
    title = self.purrer.logtitle or "Unnamed log"
    self.title_editor.setText(title);
    self.viewer_dialog.setWindowTitle(title);

  def _showViewerDialog (self):
    self._updateViewer(True);
    self.viewer_dialog.show();

  @staticmethod
  def fileModTime (path):
    try:
      return os.path.getmtime(path);
    except:
      return None;

  def _updateViewer (self,force=False):
    """Updates the viewer dialog.
    If dialog is not visible and force=False, does nothing.
    Otherwise, checks the mtime of the current purrer index.html file against self._viewer_timestamp.
    If it is newer, reloads it.
    """;
    if not force and not self.viewer_dialog.isVisible():
      return;
    # default text if nothing is found
    path = self.purrer.indexfile;
    mtime = self.fileModTime(path);
    # return if file is older than our content
    if mtime and mtime <= (self._viewer_timestamp or 0):
      return;
    busy = BusyIndicator();
    self.viewer_dialog.setDocument(path,empty=
      "<P>Nothing in the log yet. Try adding some log entries.</P>");
    self.viewer_dialog.reload();
    self.viewer_dialog.setLabel("""<P>Below is your full HTML-rendered log. Note that this is 
      only a bare-bones viewer, so only a limited set of links will work. 
      For a fully-functional view, use a proper HTML browser to look at the index file residing here:<BR>
      <tt>%s</tt></P>
      """%self.purrer.indexfile);
    self._viewer_timestamp = mtime;

  def _setEntries (self,entries):
    self.etw.clear();
    item = None;
    self._index_paths = {};
    self._index_paths[os.path.abspath(self.purrer.indexfile)] = -1;
    for i,entry in enumerate(entries):
      item = self._addEntryItem(entry,i,item);
      self._index_paths[os.path.abspath(entry.index_file)] = i;
    self.etw.resizeColumnToContents(0);

  def _titleChanged (self):
    self.setLogTitle(str(self.title_editor.text()));

  def _checkPounceStatus (self):
    ## pounce = bool([ entry for entry in self._dir_entries.itervalues() if entry.watching ]);
    pounce = bool([path for path,state in self.purrer.watchedDirectories() if state >= Purr.WATCHED ]);
    # rescan, if going from not-pounce to pounce
    if pounce and not self._pounce:
      self._rescan();
    self._pounce = pounce;
    # start timer -- we need it running to check the purr pipe, anyway
    self._timer.start(2000);

  def _forceRescan (self):
    if not self.purrer:
      self.attachDirectory('.');
    self._rescan(force=True);

  def _rescan (self,force=False):
    if not self.purrer:
      return;
    # if pounce is on, tell the Purrer to rescan directories
    if self._pounce or force:
      dps = self.purrer.rescan();
      if dps:
        filenames = [dp.filename for dp in dps];
        dprint(2,"new data products:",filenames);
        self.message("Pounced on "+", ".join(filenames));
        if self.new_entry_dialog.addDataProducts(dps):
          dprint(2,"showing dialog");
          self.new_entry_dialog.show();
    # else read stuff from pipe
    for pipe in self.purrpipes.itervalues():
      do_show = False;
      for command,show,content in pipe.read():
        if command == "title":
          self.new_entry_dialog.suggestTitle(content);
        elif command == "comment":
          self.new_entry_dialog.addComment(content);
        elif command == "pounce":
          self.new_entry_dialog.addDataProducts(self.purrer.makeDataProducts(
                                                [(content,not show)],unbanish=True));
        else:
          print "Unknown command received from Purr pipe: ",command;
          continue;
        do_show = do_show or show;
      if do_show:
        self.new_entry_dialog.show();

  def _addDPFiles (self,*files):
    """callback to add DPs corresponding to files.""";
    # quiet flag is always true
    self.new_entry_dialog.addDataProducts(self.purrer.makeDataProducts(
                          [(file,True) for file in files],unbanish=True,unignore=True));

  def _addDPFilesToOldEntry (self,*files):
    """callback to add DPs corresponding to files.""";
    # quiet flag is always true
    self.view_entry_dialog.addDataProducts(self.purrer.makeDataProducts(
                          [(file,True) for file in files],unbanish=True,unignore=True));

  def _entrySelectionChanged (self):
    selected = [ item for item in self.etw.iterator(self.etw.Iterator.Selected) if item._ientry is not None ];
    self.weditbtn.setEnabled(len(selected) == 1);
    self.wdelbtn.setEnabled(bool(selected));

  def _viewEntryItem (self,item=None,*dum):
    """Pops up the viewer dialog for the entry associated with the given item.
    If 'item' is None, looks for a selected item in the listview.
    The dum arguments are for connecting this to QTreeWidget signals such as doubleClicked().
    """;
    # if item not set, look for selected items in listview. Only 1 must be selected.
    select = True;
    if item is None:
      selected = [ item for item in self.etw.iterator(self.etw.Iterator.Selected) if item._ientry is not None ];
      if len(selected) != 1:
        return;
      item = selected[0];
      select = False; # already selected
    else:
      # make sure item is open -- the click will cause it to close
      self.etw.expandItem(item);
    # show dialog
    ientry = getattr(item,'_ientry',None);
    if ientry is not None:
      self._viewEntryNumber(ientry,select=select);

  def _viewEntryNumber (self,ientry,select=True):
    """views entry #ientry. Also selects entry in listview if select=True""";
    # pass entry to viewer dialog
    self._viewing_ientry = ientry;
    entry = self.purrer.entries[ientry];
    busy = BusyIndicator();
    self.view_entry_dialog.viewEntry(entry,
        prev=ientry>0 and self.purrer.entries[ientry-1],
        next=ientry<len(self.purrer.entries)-1 and self.purrer.entries[ientry+1]);
    self.view_entry_dialog.show();
    # select entry in listview
    if select:
      self.etw.clearSelection();
      self.etw.setItemSelected(self.etw.topLevelItem(ientry),True);

  def _viewPrevEntry (self):
    if self._viewing_ientry is not None and self._viewing_ientry > 0:
      self._viewEntryNumber(self._viewing_ientry-1);

  def _viewNextEntry (self):
    if self._viewing_ientry is not None and self._viewing_ientry < len(self.purrer.entries)-1:
      self._viewEntryNumber(self._viewing_ientry+1);
      
  def _viewPath (self,path):
    num = self._index_paths.get(os.path.abspath(path),None)
    if num is None:
      return;
    elif num == -1:
      self.view_entry_dialog.hide();
      self._showViewerDialog();
    else:
      self._viewEntryNumber(num);

  def _showItemContextMenu (self,item,point,col):
    """Callback for contextMenuRequested() signal. Pops up item menu, if defined""";
    menu = getattr(item,'_menu',None);
    if menu:
      settitle = getattr(item,'_set_menu_title',None);
      if settitle:
        settitle();
      # self._current_item tells callbacks what item the menu was referring to
      point = self.etw.mapToGlobal(point);
      self._current_item = item;
      self.etw.clearSelection();
      self.etw.setItemSelected(item,True);
      menu.exec_(point);
    else:
      self._current_item = None;

  def _copyItemToClipboard (self):
    """Callback for item menu.""";
    if self._current_item is None:
      return;
    dp = getattr(self._current_item,'_dp',None);
    if dp and dp.archived:
      path = dp.fullpath.replace(" ","\\ ");
      QApplication.clipboard().setText(path,QClipboard.Clipboard);
      QApplication.clipboard().setText(path,QClipboard.Selection);

  def _restoreItemFromArchive (self):
    """Callback for item menu.""";
    if self._current_item is None:
      return;
    dp = getattr(self._current_item,'_dp',None);
    if dp and dp.archived:
      dp.restore_from_archive(parent=self);

  def _deleteSelectedEntries (self):
    remaining_entries = [];
    del_entries = list(self.etw.iterator(self.etw.Iterator.Selected));
    remaining_entries = list(self.etw.iterator(self.etw.Iterator.Unselected));
    if not del_entries:
      return;
    hide_viewer = bool([ item for item in del_entries if self._viewing_ientry == item._ientry ]);
    del_entries = [ self.purrer.entries[self.etw.indexOfTopLevelItem(item)] for item in del_entries ];
    remaining_entries = [ self.purrer.entries[self.etw.indexOfTopLevelItem(item)] for item in remaining_entries ];
    # ask for confirmation
    if len(del_entries) == 1:
      msg = """<P><NOBR>Permanently delete the log entry</NOBR> "%s"?</P>"""%del_entries[0].title;
      if del_entries[0].dps:
        msg += """<P>%d data product(s) saved with this
                  entry will be deleted as well.</P>"""%len(del_entries[0].dps);
    else:
      msg = """<P>Permanently delete the %d selected log entries?</P>"""%len(del_entries);
      ndp = 0;
      for entry in del_entries:
        ndp += len(filter(lambda dp:not dp.ignored,entry.dps));
      if ndp:
        msg += """<P>%d data product(s) saved with these entries will be deleted as well.</P>"""%ndp;
    if QMessageBox.warning(self,"Deleting log entries",msg,
          QMessageBox.Yes,QMessageBox.No) != QMessageBox.Yes:
      return;
    if hide_viewer:
      self.view_entry_dialog.hide();
    # reset entries in purrer and in our log window
    self._setEntries(remaining_entries);
    self.purrer.deleteLogEntries(del_entries);
#    self.purrer.setLogEntries(remaining_entries);
    # log will have changed, so update the viewer
    self._updateViewer();
    # delete entry files
    for entry in del_entries:
      entry.remove_directory();

  def _addEntryItem (self,entry,number,after):
    item = entry.tw_item = QTreeWidgetItem(self.etw,after);
    timelabel = self._make_time_label(entry.timestamp);
    item.setText(0,timelabel);
    item.setText(1," "+(entry.title or ""));
    item.setToolTip(1,entry.title);
    if entry.comment:
      item.setText(2," "+entry.comment.split('\n')[0]);
      item.setToolTip(2,"<P>"+entry.comment.replace("<","&lt;").replace(">","&gt;"). \
                                    replace("\n\n","</P><P>").replace("\n","</P><P>")+"</P>");
    item._ientry = number;
    item._dp = None;
    item._menu = self._entry_menu;
    item._set_menu_title = lambda:self._entry_menu_title.setText('"%s"'%entry.title);
    # now make subitems for DPs
    subitem = None;
    for dp in entry.dps:
      if not dp.ignored:
        subitem = self._addDPSubItem(dp,item,subitem);
    self.etw.collapseItem(item);
    self.etw.header().headerDataChanged(Qt.Horizontal,0,2);
    return item;

  def _addDPSubItem (self,dp,parent,after):
    item = QTreeWidgetItem(parent,after);
    item.setText(1,dp.filename);
    item.setToolTip(1,dp.filename);
    item.setText(2,dp.comment or "");
    item.setToolTip(2,dp.comment or "");
    item._ientry = None;
    item._dp = dp;
    item._menu = self._archived_dp_menu;
    item._set_menu_title = lambda:self._archived_dp_menu_title.setText(os.path.basename(dp.filename));
    return item;

  def _make_time_label (self,timestamp):
    return time.strftime("%b %d %H:%M",time.localtime(timestamp));

  def _newLogEntry (self,entry):
    """This is called when a new log entry is created""";
    # add entry to purrer
    self.purrer.addLogEntry(entry);
    # add entry to listview if it is not an ignored entry
    # (ignored entries only carry information about DPs to be ignored)
    if not entry.ignore:
      if self.etw.topLevelItemCount():
        lastitem = self.etw.topLevelItem(self.etw.topLevelItemCount()-1);
      else:
        lastitem = None;
      self._addEntryItem(entry,len(self.purrer.entries)-1,lastitem);
      self._index_paths[os.path.abspath(entry.index_file)] = len(self.purrer.entries)-1;
    # log will have changed, so update the viewer
    if not entry.ignore:
      self._updateViewer();
      self.show();

  def _entryChanged (self,entry):
    """This is called when a log entry is changed""";
    # resave the log
    self.purrer.save();
    # redo entry item
    if entry.tw_item:
      number = entry.tw_item._ientry;
      entry.tw_item = None;
      self.etw.takeTopLevelItem(number);
      if number:
        after = self.etw.topLevelItem(number-1);
      else:
        after = None;
      self._addEntryItem(entry,number,after);
    # log will have changed, so update the viewer
    self._updateViewer();

  def _regenerateLog (self):
    if QMessageBox.question(self.viewer_dialog,"Regenerate log","""<P><NOBR>Do you really want to regenerate the
      entire</NOBR> log? This can be a time-consuming operation.</P>""",
          QMessageBox.Yes,QMessageBox.No) != QMessageBox.Yes:
      return;
    self.purrer.save(refresh=True);
    self._updateViewer();



# -*- coding: utf-8 -*-
Version = "1.1";

# some global constants

# these are the different watching modes for the directories
REMOVED = -1;     # directory removed from watchlist
UNWATCHED = 0;   # directory in list, but not watched
WATCHED = 1;        # directory watched for new files quietly
POUNCE = 2;      # directory watched loudly

import os.path

# init debug printing
import Kittens.utils
verbosity = Kittens.utils.verbosity(name="purr");
dprint = verbosity.dprint;
dprintf = verbosity.dprintf;

import Kittens.pixmaps
pixmaps = Kittens.pixmaps.PixmapCache("Purr");

import Kittens.config
ConfigFile = Kittens.config.DualConfigParser("purr.conf");
Config = Kittens.config.SectionParser(ConfigFile,"Purr");

from LogEntry import DataProduct,LogEntry
from Purrer import Purrer
import Parsers

# if GUI is enabled, this will be overwritten by an object that
# sets a busy cursor on instantiation, and restores the cursor when killed
class BusyIndicator (object):
  pass;

def canonizePath (path):
  """Returns the absolute, normalized, real path  of something.
  This takes care of symbolic links, redundant separators, etc.""";
  return os.path.abspath(os.path.normpath(os.path.realpath(path)));

def progressMessage (msg,sub=False):
  """shows a progress message. If a GUI is available, this will be redefined by the GUI.
  If sub is true, message is a sub-message of previous one.""";
  pass;



#!/usr/bin/python
# -*- coding: utf-8 -*-

# loading file into karma:
#   ksend -add_connection unix 9975 multi_array kvis -load ARRAY:FILE0 img4096.fits
# ls /tmp/.KARMA-connections for port number

import sys

def trace_lines (frame, event, arg):
  if event == "line":
    print "%s:%d"%(frame.f_code.co_filename,frame.f_lineno);
  return trace_lines;

# runs Purr standalone
if __name__ == "__main__":
  print "Welcome to PURR!"

  # parse options is the first thing we should do
  from optparse import OptionParser
  usage = "usage: %prog [options] <directories to watch>"
  parser = OptionParser(usage=usage)
  parser.add_option("-d", "--debug",dest="verbose",type="string",action="append",metavar="Context=Level",
                    help="(for debugging Python code) sets verbosity level of the named Python context. May be used multiple times.");
  parser.add_option("--trace",dest="trace",action="store_true",
                                        help="(for debugging Python code) enables line tracing of Python statements");
  (options, rem_args) = parser.parse_args();

  if options.trace:
    sys.settrace(trace_lines);

  print "Please wait a second while the GUI starts up."

  import sys
  import signal
  import os
  import os.path

  from PyQt4.Qt import *

  import Purr
  import Purr.MainWindow
  import Purr.Render
  import Purr.Startup

  import Kittens.pixmaps

  app = QApplication(sys.argv);
  app.setDesktopSettingsAware(True);

#  splash = QSplashScreen(Purr.pixmaps.purr_logo.pm());
#  splash.showMessage("PURR!");
#  splash.show();

  purrwin = Purr.MainWindow.MainWindow(None);

  try:
    if not Purr.Startup.startWizard(rem_args,purrwin):
      print "Cancelled by user";
      os._exit(1);
  except Purr.Startup.Error,err:
    print err.error_message;
    os._exit(1);

  # handle SIGINT
  def sigint_handler (sig,stackframe):
    print "Caught Ctrl+C, PURR exiting..."
    purrwin.detachPurrlog();
    app.quit();
  signal.signal(signal.SIGINT,sigint_handler);

  app.exec_();
  
  os._exit(0);



# -*- coding: utf-8 -*-
import Purr.Render

import os.path
import traceback

from Purr.Render import DefaultRenderer,dprint,dprintf


class CachingRenderer (DefaultRenderer):
  """This is a base class for "expensive" renderers.
  It maintains a cache of rendered HTML code, and only invokes the renderer if the cache is out of date.""";
  
  def _cacheFileExtension (self,cachetype,relpath):
    return ("-cache-%s-rel-%s.html"%(cachetype,bool(relpath))).lower();
  
  def __init__ (self,dp,refresh=0):
    DefaultRenderer.__init__(self,dp,refresh=refresh);
    self.rendercache = {};
    self._regenerated = False;
    
  def regenerate (self):
    """This is meant to be implemented by subclasses. Called if any cache element is out of date,
    to do the actual heavy work.""";
    pass;
  
  def checkCache (self,cachetype,relpath):
    """checks cache for rendered html code.
    'cachetype' can be any string (usually the name of a method, e.g. "Thumbnail", or "InTable")
    relpath is the relative path, here treated as boolean.
    If cache is up-to-date, returns tuple of cachekey,html (NB: cachekey is path to cache file)
    If cache is out of date, calls regenerate() if not already done so, and returns tuple of cachekey,None.
    """;
    cache_ext = self._cacheFileExtension(cachetype,relpath);
    filename,path = self.subproductPath(cache_ext);
    cachekey = path;
    # check already read cache
    content = self.rendercache.get(path,None);
    if content is not None:
      return content;
    # if cache file is up-to-date, attempt to read content
    if self.subproductUpToDate(path):
      dprintf(3,"render cache %s is up-to-date, reading in\n",path);
      try:
	content = file(path).read();
      except:
	print "Error reading render cache file",path,", will regenerate";
	traceback.print_exc();
    else:
      dprintf(3,"render cache %s is out of date, will regenerate\n",path);
    # read content? cache and return
    if content is not None:
      self.rendercache[path] = content;
      return path,content;
    # else regenerate
    if not self._regenerated:
      self.regenerate();
      self._regenerated = True;
    # and return path,None to indicate no cache
    return path,None;
    
  def writeCache (self,cachekey,content):
    """stores content in cache. 'cachekey' must be the same key as returned from checkCache().
    Returns the content."""
    self.rendercache[cachekey] = content;
    try:
      file(cachekey,'w').write(content);
    except:
      print "Error writing cache file",cachekey,", will regenerate next time";
      traceback.print_exc();
    return content;



# -*- coding: utf-8 -*-
import os.path
import os
import traceback

import Purr
from Purr import dprint,dprintf,verbosity
import Purr.Render
from Purr.Render import quote_url

from PyQt4.Qt import *


FULLINDEX = "fullindex.html";
INDEX = "index.html";

PURR_ICONS =  { 16:"purr16.png", 24:"purr24.png",32:"purr32.png" };

def initIndexDir (logdir):
  for sz,filename in PURR_ICONS.iteritems():
    path = os.path.join(logdir, filename);
    if not os.path.exists(path):
      Purr.pixmaps.purr_logo.pm().scaled(QSize(sz,sz),Qt.KeepAspectRatioByExpanding,Qt.SmoothTransformation).save(path);

def renderIcon (size,path=None):
  filename = PURR_ICONS[size];
  if path:
    filename = os.path.join(path,filename);
  return """<IMG SRC="%s"></IMG>"""%quote_url(filename);

def writeLogIndex (logdir,title,timestamp,entries,refresh=0):
  logdir = os.path.normpath(os.path.abspath(logdir));
  fullindex = os.path.join(logdir,FULLINDEX);
  tocindex = os.path.join(logdir,INDEX);

  # if 5 entries or less, do an index only, and remove the fullindex
  if len(entries) <= 5:
    if os.path.exists(fullindex):
      try:
        os.remove(fullindex);
      except:
        print "Error removing %s:"%fullindex;
        traceback.print_exc();
    fullindex = tocindex;
    tocindex = None;

  # generate TOC index, if asked to
  if tocindex:
    fobj = file(tocindex, "wt");
    # write index.html
    fobj.write("""<HTML><BODY>\n
      <TITLE>%s</TITLE>

      <H1><A CLASS="TITLE" TIMESTAMP=%.6f>%s</A></H1>

      """%(title,timestamp,title));
    fobj.write("""<DIV ALIGN=right><P><A HREF=%s>Printable version (single HTML page).</A></P></DIV>\n\n"""%FULLINDEX);
    # write entries
    for i,entry in enumerate(entries):
      path = os.path.normpath(os.path.abspath(entry.index_file));
      if path.startswith(logdir+'/'):
        path = path[(len(logdir)+1):];
      fobj.write("""<P><A HREF="%s">%d. %s</A></P>\n\n"""%(quote_url(path),i+1,entry.title));
    # write footer
    fobj.write("<HR>\n\n");
    fobj.write("""<DIV ALIGN=right>%s <I><SMALL>This log was generated
               by PURR version %s.</SMALL></I></DIV>\n"""%(renderIcon(16),Purr.Version));
    fobj.write("</BODY></HTML>\n");
    fobj = None;

  # generate full index, if asked to
  if fullindex:
    fobj = file(fullindex, "wt");
    # write index.html
    fobj.write("""<HTML><BODY>\n
      <TITLE>%s</TITLE>

      <H1><A CLASS="TITLE" TIMESTAMP=%.6f>%s</A></H1>

      """%(title,timestamp,title));
    # write entries
    for entry in entries:
      fobj.write(entry.renderIndex(os.path.join(os.path.basename(entry.pathname),""),refresh=refresh));
    # write footer
    fobj.write("<HR>\n");
    fobj.write("""<DIV ALIGN=right>%s <I><SMALL>This log was generated
               by PURR version %s.</SMALL></I></DIV>\n"""%(renderIcon(16),Purr.Version));
    fobj.write("</BODY></HTML>\n");
    fobj = None;



"""Purr.Pipe is a very lightweight module for writing to Purr pipes.
It is meant to be used within TDL scripts and such."""

import os.path
import traceback

_pipefile = ".purr.pipe";

class Pipe (object):
  """The Pipe class represents a Purr pipe.""";
  def __init__ (self,dirname):
    """opens a Purr pipe to the specified directory""";
    self.dirname = dirname;
    global _pipefile;
    file = os.path.join(dirname,_pipefile);
    self.pipefile = os.path.abspath(os.path.normpath(os.path.realpath(file)));
    self.pipefile_read = os.path.abspath(os.path.normpath(os.path.realpath(file)))+".read";
    self.mtime = 0;
    
  # writer section
  
  def _write (self,what):
    """writes something to the Purr pipe""";
    try:
      file(self.pipefile,"a").write(what);
    except:
      print "Error writing to %s:"%self.pipefile;
      traceback.print_exc();
  
  def title (self,title,show=False):
    """writes a title tag to the Purr pipe""";
    self._write("title:%d:%s\n"%(int(show),title));
    return self;
  
  def comment (self,comment,show=False,endline=True,newpar=False):
    """writes a comment tag to the Purr pipe""";
    if not endline:
      comment += "<NOBR>";
    if newpar:
      comment += "<BR>";
    self._write("comment:%d:%s\n"%(int(show),comment));
    return self;
    
  def __lshift__ (self,comment):
    """<< is equivalent to writing a comment""";
    return self.comment(comment);
    
  def pounce (self,file,show=False):
    """writes a pounce command to the Purr pipe""";
    file = os.path.abspath(os.path.normpath(os.path.realpath(file)));
    self._write("pounce:%d:%s\n"%(int(show),file));
    return self;
    
  # reader section
  def read (self):
    """Checks if anything has arrived in the pipe file, returns list of 
    (command,show,content) tuples.
    Special tags added in comment() above are replaced.
    """;
    # read file is set to None on error, so that we stop reading
    if self.pipefile_read is None:
      return [];
    # no pipe -- read nothing
    if not os.path.exists(self.pipefile):
      return [];
    # move pipe file out of the way before we start reading from it
    try:
      os.rename(self.pipefile,self.pipefile_read);
    except:
      print "Error renaming %s to %s:"%self.pipefile,self.pipefile_read;
      traceback.print_exc();
      self.pipefile_read = None;
      return [];
    # now read everything
    try:
      lines = file(self.pipefile_read).readlines();
    except:
      print "Error reading %s:"%self.pipefile_read;
      traceback.print_exc();
      self.pipefile_read = None;
      return [];
    # parse lines 
    cmds = [];
    for line in lines:
      tokens = line.rstrip().split(":",2);
      if len(tokens) < 3:
        continue;
      command,show,content = tokens;
      # convert 'show' to bool
      try:
        show = bool(int(show));
      except:
        show = False;
      # if 'comment' command, parse content string
      if command == "comment":
        # <BR> inserts new paragraph (double-newline)
        content = content.replace("<BR>","\n\n");
        # <NOBR> blocks newline at end, which is otherwise implictly added...
        if content.endswith("<NOBR>"):
          content = content[:-6];
        # ...here
        elif not content.endswith("\n"):
          content += "\n";
      # add to list of contents
      # print "read %s:%d:%s:"%(command,show,content);
      cmds.append((command,show,content));
    return cmds;
  
def open (dirname):
  return Pipe(dirname);




#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import pychart_util
import types
AnyType = 9998

def IntervalType(val):
    if type(val) in (types.IntType, types.LongType,
                     types.FloatType, types.FunctionType):
	return None
    return "Expecting a number or a function"

def CoordType(val):
    if type(val) not in (types.TupleType, types.ListType):
        return "Expecting a tuple or a list"
    if len(val) != 2:
        return "Coordinate must be a pair of numbers"
    for v in val:
        if v != None and NumberType(v):
            return "Expecting a pair of numbers (got %s)" % str(v)
    return None 
    
def NumberType(val):
    if type(val) in (types.IntType, types.LongType, types.FloatType):
        return None
    else:
        return "Expecting a number"

def UnitType(val):
    if type(val) in (types.IntType, types.LongType, types.FloatType):
        return
    else:
        return "Expecting a unit"
def ShadowType(val):
    if type(val) not in (types.TupleType, types.ListType):
	return "Expecting tuple or list."
    if len(val) != 3:
	return "Expecting (xoff, yoff, fill)."
    return None

def FormatType(val):
    if type(val) in (types.StringType, types.FunctionType):
        return None
    return "Format must be a string or a function"

